# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__all__ = [
    'NeuralGraph',
]

from collections import OrderedDict, defaultdict, namedtuple
from os import path
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from ruamel.yaml import YAML
from torch.nn.parallel import DataParallel

from nemo.backends import get_state_dict, load, save, set_state_dict
from nemo.core.neural_factory import DeviceType, OperationMode
from nemo.core.neural_interface import NeuralInterface
from nemo.core.neural_modules import ModuleType, NeuralModule
from nemo.core.neural_types import NeuralPortNameMismatchError, NeuralType, NmTensor
from nemo.package_info import __version__ as nemo_version
from nemo.utils import logging
from nemo.utils.configuration_error import ConfigurationError
from nemo.utils.neural_graph.connection import Connection, StepModulePort
from nemo.utils.neural_graph.graph_inputs import GraphInputs
from nemo.utils.neural_graph.graph_outputs import GraphOutputs

YAML = YAML(typ='safe')


class NeuralGraph(NeuralInterface):
    """
    Neural Graph class stores dynamically defined graphs of connected Neural Modules.
    Once defined, the graph is static, i.e. the connectivity between the nodes is frozen.
    """

    def __init__(self, operation_mode: OperationMode = OperationMode.both, name: Optional[str] = None):
        """
        Constructor. Initializes graph variables.

        Args:
            operation_mode: Graph operation mode, that will be propagated along modules during graph creation.
            [training | eval | both] (DEFAULT: both)
            name: Name of the graph (optional)
        """
        # Initialize the inferface.
        super().__init__()

        # Register graph.
        self._name = self._app_state.register_graph(self, name)

        # Store name and operation mode.
        self._operation_mode = operation_mode

        # "Modules" - list of modules constituting "nodes" in a given graph.
        self._modules = {}

        # All tensors produced within this graph (dict of dicts).
        # This stores  "all output tensors" dictionary, where the key is the name of "producer" module,
        # and the value contains a dictionary of all tensors produced by it.
        self._all_tensors = {}

        # "Steps": order of the  execution of modules in a graph.
        self._steps = OrderedDict()

        # Bound inputs.
        self._inputs = GraphInputs()

        # Bound outputs.
        self._outputs = GraphOutputs(self._all_tensors)

        # Flag indicating whether the "default" output ports/tensors will be automatically bound.
        self.default_output_binding = True

        # Set default data loader params.
        self._data_loader_config = {
            "dataset": None,
            "batch_size": 1,
            "shuffle": False,
            "sampler": None,
            "batch_sampler": None,
            "num_workers": 0,
            "collate_fn": None,
            "pin_memory": False,
            "drop_last": False,
            "timeout": 0,
            "worker_init_fn": None,
            "multiprocessing_context": None,
        }

        # Lazy-initialize loader when needed.
        self._data_loader = None

        # Data collected during forward propagation.
        self._forward_data = defaultdict(lambda: {})

        # Initial device: CPU.
        self._pt_device = torch.device("cpu")

    def __call__(self, **kwargs):
        """
        This method "nests" one existing neural graph into another one.
        Also checks if all inputs were provided and properly connects them.

        Args:
            kwargs: keyword arguments containing dictionary of (input_port_name, port_content).
        """
        # Test operation modes of the nested graphs.
        outer_mode = self._app_state.active_graph.operation_mode
        inner_mode = self.operation_mode

        if inner_mode == OperationMode.evaluation and outer_mode == OperationMode.training:
            raise TypeError("Cannot nest 'inference' graph into 'training'")

        if inner_mode == OperationMode.training and outer_mode == OperationMode.evaluation:
            raise TypeError("Cannot nest 'training' graph into 'inference'")

        if inner_mode == OperationMode.training and outer_mode == OperationMode.both:
            raise TypeError("Cannot nest 'training' graph into 'both'")

        if inner_mode == OperationMode.evaluation and outer_mode == OperationMode.both:
            raise TypeError("Cannot nest 'inference' graph into 'both'")

        # Check inputs: iterate through all inputs passed to the "self".
        for port_name, port_content in kwargs.items():
            # Make sure that passed arguments correspond to input port names.
            if port_name not in self.input_ports.keys():
                raise NeuralPortNameMismatchError(port_name)

        # "Nest" this graph into an active graph.
        results = self._app_state.active_graph.__nest(self, kwargs)

        # Return output tensors.
        return results

    def __nest(self, inner_graph: 'NeuralGraph', inner_graph_args):
        """
        Method nests (copies) a graph: modules, steps, topology (tensors).

        Args:
            inner_graph: Graph to be copied (will be "nested" in this (self) graph).
            inner_graph_args: inputs passed to the graph call.
        """
        # Remember the number of "already present steps".
        step_bump = len(self.steps)

        # "Copy" the modules from nested graph.
        for key, module in inner_graph.modules.items():
            # Check if module with that name already exists.
            # TODO: Uncomment when we will refactor all examples so training/validation graphs won't be added
            # to the "default" graph.
            # if key in self._modules.keys():
            #    raise KeyError("Neural Graph already contains a module named {}".format(module.name))
            self._modules[key] = module

        # Next we should copy the topography - i.e. produce "real" copies of tensors.
        # In fact, instead of copying, we will produce them, following:
        # - the execution order defined in "steps"
        # - connectivity defined in tensor' consumers-ports
        # (so the same logic that will be used in graph deserialization)

        # So let us first serialize the connections of the nested graph.
        # Create a list: (producer.port -> consumer.port)
        inner_connections = []
        for tensors in inner_graph.tensors.values():
            for t in tensors.values():
                inner_connections.extend(t.connections())

        # We need to disable the binding of "defeault" ports on per module basis - we will "manually" produce
        # them only for ports that are already indicated as the "bound" ones in the inner graph.
        self.default_output_binding = False

        # Now "copy" graph execution order and topology by actually executing each step of the nested graph.
        for step_number, module_name in inner_graph.steps.items():
            # Both module and step will be added by the modules' call().

            # Get the module.
            module = inner_graph._modules[module_name]

            # Produce list of arguments that will be passed to a given modules.
            module_args = {}
            # Do it by:
            # - harvesing input port names of a given module,
            # - checking if the input was not bound (in the inner graph),
            # - checking if we have already tensors leading to that input (in outer graph).
            for input_port_name in module.input_ports.keys():
                # Check if this port was bound in the inner graph.
                key = inner_graph.inputs.has_binding(step_number, input_port_name)
                # If so, then we must pass whatever was passed to that port in the list of arguments.
                if key is not None:
                    module_args[input_port_name] = inner_graph_args[key]
                    # As a result, the "module" call() will bind this input!
                    continue

                # Else: find a tensor that should be passed to the given module's input.
                # Search for producer/port that we should use.
                for connection in inner_connections:
                    if (
                        connection.consumer.step_number == step_number
                        and connection.consumer.module_name == module_name
                        and connection.consumer.port_name == input_port_name
                    ):
                        # Got the connection!
                        bumped_step = connection.producer.step_number + step_bump
                        # producer_name = connection.producer.module_name
                        producer_port_name = connection.producer.port_name
                        break

                # Now, the tensor is already produced in outer (i.e. this) graph!
                module_args[input_port_name] = self.tensors[bumped_step][producer_port_name]

            # Ok, now we have all keyword arguments. We can call() the module.
            # This will collect all the produced output tensors and add them to this graph.
            module(**module_args)

        # At that point we have all modules, steps and tensors added to outer (self) graph.
        # Now we have to prepare the outputs.

        # This part is different from Neural Module.
        # Now the goal is NOT to create NEW "tensors", but to return the BOUND ones!
        # Still, those must be bound in the outer (active) graph, but using port names from the inner (nested) graph.

        # Get list of "the adequate output tensors".
        output_tensors = {}
        # Iterate through outputs of the inner graph.
        for key, tensor in inner_graph.output_tensors.items():
            # Find the tensors within this (outer) graph that are outputs by the same producer-port.
            bumped_step = tensor.producer_step_number + step_bump
            # producer_name = tensor.producer_name
            producer_port_name = tensor.name
            # Get adequate tensor from "outer graph" (self).
            output_tensors[key] = self.tensors[bumped_step][producer_port_name]

        if len(output_tensors) == 1:
            # Return a single tensor.
            key = list(output_tensors)[0]
            results = output_tensors[key]

            # Bind the "default" output ports of the inner graph as "default" output ports of this graph.
            # Call the bind() method of bound_outputs directly, as we already have the tensors in our graph.
            # But: Use output port name of the inner graph!
            self.outputs.bind([results], [key])

        else:
            # Create a named tuple type enabling to access outputs by attributes (e.g. out.x).
            output_class_name = f'{self.__class__.__name__}Output'
            result_type = namedtuple(typename=output_class_name, field_names=output_tensors.keys())

            # Return the bound output tensors.
            results = result_type(*output_tensors.values())

            # Bind the "default" output ports of the inner graph as "default" output ports of this graph.
            # Call the bind() method of bound_outputs directly, as we already have the tensors in our graph.
            # But: Use output port name of the inner graph!
            self.outputs.bind(output_tensors.values(), output_tensors.keys())

        # Ok, now we can turn automatic binding on.
        self.default_output_binding = True

        # Return the results.
        return results

    def record_step(self, module: NeuralModule):
        """
        Records the operation (the module to be executed) on a list.

        Args:
            module: Neural modules added to a given graph.

        Returns:
            Step number.
        """
        # The solution allows loops in the graph.
        # This also means that module with that name can already be present in the graph.
        if module.name in self._modules.keys():
            # Check if this is the same module.
            if self._modules[module.name] is not module:
                raise KeyError("Neural Graph already contains a different module with name `{}`!".format(module.name))

        else:
            # Add module to list of modules.
            self._modules[module.name] = module

        # Add step - store the module name.
        step_number = len(self._steps)
        self._steps[step_number] = module.name

        # Return the current step number.
        return step_number

    @property
    def step_number(self) -> int:
        """
        Returns:
            The current step number.
        """
        return len(self._steps) - 1

    def bind_outputs(self, tensors_list: Union[NmTensor, List[NmTensor]]):
        """
        Binds the output tensors.

        Args:
            tensors_list: A single tensor OR a List of tensors to be bound.
        """
        # Handle both single port and lists of ports to be bound.
        if type(tensors_list) is not list:
            tensors_list = [tensors_list]

        # Add tensors to list of list of tensors.
        for tensor in tensors_list:
            # Add tensor to "all" tensors dictionary.
            step_number = tensor.producer_step_number
            if step_number not in self._all_tensors.keys():
                self._all_tensors[step_number] = {}

            port_name = tensor.name
            # Add tensor.
            self._all_tensors[step_number][port_name] = tensor

        # Bind the tensors as graph outputs.
        if self.default_output_binding:
            self.outputs.bind(tensors_list)

    @property
    def inputs(self) -> GraphInputs:
        """
        Returns:
            Graph input.
        """
        return self._inputs

    @property
    def input_ports(self) -> Dict[str, NeuralType]:
        """
        Returns definitions of graph input ports (dict of Neural Types).

        .. note::
            This method actually returns an immutable  dictionary with port types (like Neural Modules).
            In order to get access to actual graph inputs please call the inputs() method.

        Returns:
            Graph input ports definitions.
        """
        return self._inputs.definitions

    @property
    def outputs(self) -> GraphOutputs:
        """
        Returns graph outputs.

        Returns:
            Graph outputs.
        """
        return self._outputs

    @property
    def output_ports(self) -> Dict[str, NeuralType]:
        """
        Returns definitions of module output ports (dict of Neural Types).

        .. note::
            This method actually returns an immutable dictionary with port types (like Neural Modules).
            In order to get access to actual graph outpus please call the outputs() method.

        Returns:
            Graph output ports definitions.
            
        """
        return self._outputs.definitions

    @property
    def output_tensors(self) -> Dict[str, NmTensor]:
        """
        Returns:
            Fraph output tensors.
        """
        return self._outputs.tensors

    @property
    def modules(self) -> Dict[str, NeuralModule]:
        """ Returns modules. """
        return self._modules

    def __getitem__(self, key) -> NeuralModule:
        """ Returns module given its name (name of the variable).

            Args:
                key: Name of the variable.
            
            Raises:
                KeyError: Neural Graph doesn't contain a module with a given name (key).
        """
        if key not in self._modules.keys():
            raise KeyError("Neural Graph doesn't contain a module named {}".format(key))
        return self._modules[key]

    def __len__(self) -> int:
        """
        Returns:
            The number of modules (vertices) in a given graph.
        """
        return len(self._modules)

    @property
    def steps(self) -> Dict[int, str]:
        """
        Returns:
            Dictionary [steps_number, module_name]
        """
        return self._steps

    @property
    def tensors(self):
        """
        Property returning a (double) dictionary of all output tensors.

        Returns:
            Dictionary of tensors in the format [module_name][output_port_name].
         """
        return self._all_tensors

    @property
    def tensor_list(self) -> List[NmTensor]:
        """
        Property returning output tensors by extracting them on the fly from the bound outputs.

        Returns:
            List of tensors.
        """
        tensor_list = []
        # Get tensors by acessing the producer-ports.
        for tensors_per_module in self._all_tensors.values():
            for tensor in tensors_per_module.values():
                # Add it to the list.
                tensor_list.append(tensor)
        # Return the result.
        return tensor_list

    @property
    def operation_mode(self) -> OperationMode:
        """
        Returns:
            Operation mode.
        """
        return self._operation_mode

    def __enter__(self) -> 'NeuralGraph':
        """ 
        Activates this graph.

        Returns:
            The graph object.
        """
        self._app_state.active_graph = self
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Deactivates the current graph.
        """
        self._app_state.active_graph = None

    def activate(self):
        """ 
        Activates this graph.
        """
        self._app_state.active_graph = self

    def deactivate(self):
        """
        Deactivates the current graph.
        """
        self._app_state.active_graph = None

    def export_to_config(self, config_file: str):
        """
        Exports the neural graph to a file.

        Args:
            config_file: Name (and path) of the config file (YML) to be written to.
        """
        # Greate an absolute path.
        abs_path_file = path.expanduser(config_file)

        # Serialize the graph.
        to_export = self.serialize()

        # All parameters are ok, let's export.
        with open(abs_path_file, 'w') as outfile:
            YAML.dump(to_export, outfile)

        logging.info(
            "Configuration of graph `{}` ({}) exported to '{}'".format(self.name, type(self).__name__, abs_path_file)
        )

    def serialize(self) -> Dict[str, Any]:
        """
        Method serializes the whole graph.

        Returns:
            Dictionary containing description of the whole graph.
        """
        # Create a dictionary representing the serialized object.
        serialized_graph = {}

        # Add "header" with module "specification".
        serialized_graph["header"] = self.__serialize_header()

        # Add modules.
        serialized_graph["modules"] = self.__serialize_modules()

        # Add steps.
        serialized_graph["steps"] = self.__serialize_steps()

        # Add connectinos.
        serialized_graph["connections"] = self.__serialize_connections()

        # Serialize graph (bound) inputs.
        serialized_graph["inputs"] = self._inputs.serialize()

        # Serialize graph (bound) outputs.
        serialized_graph["outputs"] = self._outputs.serialize()

        # Return the dictionary.
        return serialized_graph

    def __serialize_header(self) -> Dict[str, Any]:
        """
        Private method responsible for serializing the graph header.

        Returns:
            Dictionary containing description of the whole graph.
        """
        # Generate full_spec of the class.
        full_spec = str(self.__module__) + "." + str(self.__class__.__qualname__)
        header = {"nemo_core_version": nemo_version, "full_spec": full_spec}
        # Add operation mode.
        if self._operation_mode == OperationMode.training:
            header["operation_mode"] = "training"
        elif self._operation_mode == OperationMode.evaluation:
            header["operation_mode"] = "inference"
        else:
            header["operation_mode"] = "both"
        # Return header.
        return header

    def __serialize_modules(self) -> Dict[str, Any]:
        """
        Private method responsible for serializing the modules present in the graph.

        Returns:
            Dictionary containing description of all graph modules.
        """
        serialized_modules = {}
        for name, module in self._modules.items():
            serialized_modules[name] = module.serialize()
        return serialized_modules

    def __serialize_steps(self):
        """
        Private method responsible for serializing the steps (order of module executions).

        Returns:
            Dictionary containing description of the steps.
        """
        serialized_steps = {}
        for no, module_name in self._steps.items():
            serialized_steps[no] = module_name
        return serialized_steps

    def __serialize_connections(self) -> Dict[str, Any]:
        """
        Private method responsible for serializing the connections in the graph.

        Returns:
            List containing "connections" between modules.
        """
        serialized_connections = []
        # Iterate through "tensor modules".
        for tensors in self._all_tensors.values():
            # Iterate through "tensor output ports".
            for tensor in tensors.values():
                # "Transform" tensor to the list of connections.
                for c in tensor.connections():
                    # Serialize!
                    source = str(c.producer.step_number) + "." + c.producer.module_name + "." + c.producer.port_name
                    target = str(c.consumer.step_number) + "." + c.consumer.module_name + "." + c.consumer.port_name
                    ntype_str = str(tensor.ntype)
                    serialized_connections.append(source + "->" + target + " | " + ntype_str)
        return serialized_connections

    @classmethod
    def import_from_config(
        cls,
        config_file: str,
        reuse_existing_modules: bool = False,
        overwrite_params: Dict[str, Any] = {},
        name: Optional[str] = None,
    ) -> 'NeuralGraph':
        """
        Class method importing the neural graph from the configuration file.
        Raises an ImportError exception when config file is invalid.

        Args:
            config_file: path (absolute or relative) and name of the config file (YML)
            reuse_existing_modules: If the modules with (name, type, init_params) are already created, import will
            connect to them instead of creating new instances.
            overwrite_params: Dictionary containing parameters that will be added to or overwrite (!) the default
            parameters loaded from the configuration file
            name: Name of the new graph (optional, DEFAULT: NONE)
        Returns:
            Instance of the created NeuralGraph object.
        """
        logging.info("Loading configuration of a new Neural Graph from the `{}` file".format(config_file))

        # Validate the content of the configuration file (its header).
        loaded_config = cls.__validate_config_file(config_file)
        # TODO: overwrite params?

        # "Deserialize" the graph.
        new_graph = cls.deserialize(loaded_config, reuse_existing_modules, name)

        # Return the object.
        return new_graph

    @classmethod
    def __validate_config_file(cls, config_file: str):
        """
        Class method validating whether the config file has a proper content (sections, specification etc.).
        Raises an ImportError exception when config file is invalid or
        incompatible (when called from a particular class).

        Args:
            config_file: path (absolute or relative) and name of the config file (YML)
        Returns:
            A loaded configuration file (dictionary).
        """
        # Greate an absolute path.
        abs_path_file = path.expanduser(config_file)

        # Open the config file.
        with open(abs_path_file, 'r') as stream:
            loaded_config = YAML.load(stream)

        # Check sections.
        for section_name in ["header", "modules", "steps", "connections", "inputs", "outputs"]:
            if section_name not in loaded_config.keys():
                raise ImportError(
                    "The loaded config `{}` doesn't contain the required `{}` section".format(
                        config_file, section_name
                    )
                )

        # Parse the "full specification".
        spec_list = loaded_config["header"]["full_spec"].split(".")

        # Check if config contains definition of Neural Graph.
        if spec_list[-1] != "NeuralGraph":
            txt = "The loaded file `{}` contains configuration of ".format(config_file)
            txt = txt + "`{}` thus cannot be used for instantiation of Neural Graph".format(spec_list[-1])
            raise ImportError(txt)

        # Success - return the loaded configuration.
        return loaded_config

    @classmethod
    def deserialize(
        cls, configuration: Dict[str, Any], reuse_existing_modules: bool = False, name: Optional[str] = None
    ) -> 'NeuralGraph':
        """
        Class method creating a graph instance by deserializing the provided configuratino.

        Args:
            configuration: Dictionary containing serialized graph.
            reuse_existing_modules: If the modules with (name, type, init_params) are already created, import will
            connect to them instead of creating new instances.
        Returns:
            Instance of the created NeuralGraph object.
        """
        # Deserialize header and get object class.
        operation_mode = cls.__deserialize_header(configuration["header"])

        # Create the graph instance.
        new_graph = NeuralGraph(operation_mode=operation_mode, name=name)
        logging.info(
            "Instantiated a new Neural Graph named `{}` with mode `{}`".format(
                new_graph.name, new_graph.operation_mode
            )
        )
        # Deserialize modules.
        modules = new_graph.__deserialize_modules(configuration["modules"], reuse_existing_modules)

        # Deserialize steps.
        steps = new_graph.__deserialize_steps(configuration["steps"])

        # Deserialize the connections between modules.
        connections = new_graph.__deserialize_connections(configuration["connections"], modules)

        # Deserialize input bindings - return it in an external entity.
        inputs = GraphInputs.deserialize(configuration["inputs"], modules)

        # Deserialize "manual" output bindings.
        new_graph._outputs.deserialize(configuration["outputs"], modules)

        # Now we have to execute the graph, following the steps and connections.
        new_graph.__execute_and_create_tensors(steps, modules, connections, inputs)

        # Return the graph instance.
        return new_graph

    @classmethod
    def __deserialize_header(cls, serialized_header: Dict[str, Any]):
        """
        Private class method deserializing the header and extracts the general information.

        Args:
            serialized_header: Dictionary containing graph header.
        Returns:
            Operation mode.
        """
        # Parse the "full specification" - do not need that now.
        # spec_list = serialized_header["full_spec"].split(".")

        # Get operation mode.
        if serialized_header["operation_mode"] == "training":
            operation_mode = OperationMode.training
        elif serialized_header["operation_mode"] == "inference":
            operation_mode = OperationMode.evaluation
        else:
            operation_mode = OperationMode.both

        # Return the mode.
        return operation_mode

    def __deserialize_modules(self, serialized_modules: Dict[str, Any], reuse_existing_modules: bool):
        """
        Private method deserializing the modules present in the graph.

        Args:
            serialized_modules: Dictionary containing graph modules.
            reuse_existing_modules: If True, won create a new module when a module with a given name exists.

        Returns:
            Dictionary of modules.

        Raises:
            KeyError: A module with name already exists (if reuse_existing_modules is set to False).
        """
        modules = {}
        for name, module_params in serialized_modules.items():
            # Check if module already exists.
            if self._app_state.modules.has(name):
                # Check if we can reuse the existing modules.
                if reuse_existing_modules:
                    modules[name] = self._app_state.modules[name]
                else:
                    raise KeyError("A module with name `{}` already exists!".format(name))
            else:
                # Ok, create a new module.
                modules[name] = NeuralModule.deserialize(module_params)
        # Ok, done.
        return modules

    def __deserialize_steps(self, serialized_steps: Dict[str, Any]):
        """
        Private method deserializing the steps (order of module executions).

        Args:
            serialized_steps: Dictionary containing serialized steps.
        Returns:
            Odered dict with steps.
        """
        steps = OrderedDict()
        for i in range(len(serialized_steps)):
            steps[i] = serialized_steps[i]
        # Ok, done.
        return steps

    def __deserialize_connections(self, serialized_connections: Dict[str, Any], modules: Dict[str, NeuralModule]):
        """
        Private method deserializing the connections in the graph.

        Args:
            serialized_steps: Dictionary containing serialized connections.
            modules: List of modules.
        Returns:
            List of connections, in a format enabling graph traversing.
        """
        connections = []
        # Deserialize connections one by one.
        for c in serialized_connections:
            # Deserialize!
            [producer, consumer_type] = c.split("->")
            [consumer, ntype_str] = consumer_type.split(" | ")
            [producer_step, producer_name, producer_port_name] = producer.split(".")
            [consumer_step, consumer_name, consumer_port_name] = consumer.split(".")
            producer_mp = StepModulePort(int(producer_step), producer_name, producer_port_name)
            consumer_mp = StepModulePort(int(consumer_step), consumer_name, consumer_port_name)
            # Get tensor type.
            ntype = modules[producer_name].output_ports[producer_port_name]
            # Validate if neural type is ok.
            assert ntype_str == str(ntype)

            # Add connection.
            connections.append(Connection(producer_mp, consumer_mp, ntype))
        # Ok, done.
        return connections

    def __execute_and_create_tensors(self, steps, modules, connections, inputs):
        """
        Method creates (internal) tensors of the graph by executing it following the order and using
        the provided connections and inputs.

        Args:
            steps: List of steps to be executed.
            modules: List of modules.
            connections: List of connections.
            inputs: List of "bound inputs"
        """
        # Activate this graph, so all the tensors will be added to this !
        self.activate()

        # We need to disable the binding of "defeault" ports on per module basis.
        # We will "manually" produce (e.g. deserialize) them outside of this function.
        self.default_output_binding = False

        # Now "copy" graph execution order and topology by actually executing each step of the nested graph.
        for step, module_name in steps.items():
            # Both module and step will be added by the modules' call().

            # Get the module.
            module = modules[module_name]

            # Produce list of arguments that will be passed to a given module.
            module_args = {}
            # Do it by:
            # - harvesing input port names of a given module,
            # - checking if the input was not bound (in the inner graph),
            # - checking if we have already tensors leading to that input (in outer graph).
            for input_port_name in module.input_ports.keys():
                # Check if this port was bound in the inner graph.
                key = inputs.has_binding(step, input_port_name)

                # import pdb;pdb.set_trace()
                # If so, then we must pass the binding!
                if key is not None:
                    # Copy the port "definition" (i.e. is NeuralType) using the same port name.
                    self.inputs[key] = inputs[key]

                    # Pass this object to module input argument.
                    module_args[input_port_name] = self.inputs[key]

                # Else: find a tensor that should be passed to the given module's input.
                else:
                    # Search for producer/port that we should use.
                    for connection in connections:
                        if (
                            connection.consumer.step_number == step
                            and connection.consumer.module_name == module_name
                            and connection.consumer.port_name == input_port_name
                        ):
                            # Got the connection!
                            producer_step_number = connection.producer.step_number
                            # producer_name = connection.producer.module_name
                            producer_port_name = connection.producer.port_name
                            break
                    # Now, the tensor is already produced in outer (i.e. this) graph!
                    module_args[input_port_name] = self.tensors[producer_step_number][producer_port_name]
                # End: for

            # Ok, now we have all keyword arguments. We can call() the module.
            # This will collect all the produced output tensors and add them to this graph.
            module(**module_args)

        # At that point we have all modules, steps and tensors added to outer (self) graph.
        # Now we have to prepare the outputs.

        # Deactivate graph.
        self.deactivate()

        # Ok, now we can turn automatic binding on.
        self.default_output_binding = True

    def summary(self) -> str:
        """ 
        Returns:
            A nice, full graph summary.
        """
        # Line "decorator".
        desc = "\n" + 113 * '=' + "\n"
        # 1. general information.
        desc += "The `{}` Neural Graph [{}, ".format(self.name, self.operation_mode)
        if self.is_complete:
            desc += " COMPLETE,"
        else:
            desc += " INCOMPLETE,"
        if self.is_trainable:
            desc += " TRAINABLE]:\n"
        else:
            desc += " NON-TRAINABLE]:\n"

        # 2. modules.
        desc += " * Modules ({}):\n".format(len(self._modules))
        for key, module in self._modules.items():
            if module.type == ModuleType.trainable and module.is_frozen():
                desc += "    * `{}` ({}) [FROZEN]\n".format(key, type(module).__name__)
            else:
                desc += "    * `{}` ({})\n".format(key, type(module).__name__)

        # 3. steps.
        desc += " * Steps ({}):\n".format(len(self._steps))
        for num, module in self._steps.items():
            desc += "    {}. {}\n".format(num, module)

        # 4. connections.
        connections = self.__serialize_connections()
        desc += " * Connections ({}):\n".format(len(connections))
        # if len(connections) == 0:
        #    desc += "    -\n"
        for connection in connections:
            desc += "    * {}\n".format(connection)

        # 5. graph (bound) inputs.
        inputs = self._inputs.serialize()
        desc += " * Graph Inputs ({}):\n".format(len(inputs))
        # if len(inputs) == 0:
        #    desc += "    -\n"
        for input in inputs:
            desc += "    * {}\n".format(input)

        # 6. graph (bound) outputs.
        outputs = self._outputs.serialize()
        desc += " * Graph Outputs ({}, {}):\n".format(len(outputs["mappings"]), outputs["type"])
        # if len(outputs) == 0:
        #    desc += "    -\n"
        for output in outputs["mappings"]:
            desc += "    * {}\n".format(output)
        # Line "decorator".
        desc += 113 * '='

        # Return the result.
        return desc

    def freeze(self, module_names: Optional[List[str]] = None):
        """
        A method that freezes the weights of the trainable modules in a graph.

        Args:
            module_names: List of modules to be frozen (Optional). If passed, all modules will be unfrozen.
        Raises:
            KeyError: If name of the module won't be recognized.
        """
        # Work on all modules.
        if module_names is None:
            module_names = self._modules.keys()

        # Iterate through modules one by one.
        for name in module_names:
            if name not in self._modules.keys():
                raise KeyError("Module `{}` not present in the `{}` graph".format(name, self.name))
            # Check module type.
            module = self._modules[name]
            if module.type == ModuleType.trainable:
                # Freeze weights of the module.
                module.freeze()
            else:
                logging.debug("Module `{}` is not trainable so cannot be frozen".format(name))

    def unfreeze(self, module_names: Optional[List[str]] = None):
        """
        Unfreezes weights of the trainable modules in a graph.

        Args:
            module_names: List of modules to be unfrozen (Optional). If not passed, all modules will be unfrozen.
        Raises:
            KeyError: If name of the module won't be recognized.
        """
        # Work on all modules.
        if module_names is None:
            module_names = self._modules.keys()

        # Iterate through modules one by one.
        for name in module_names:
            if name not in self._modules.keys():
                raise KeyError("Module `{}` not present in the `{}` graph".format(name, self.name))
            # Check module type.
            module = self._modules[name]
            if module.type == ModuleType.trainable:
                # Unfreeze weights of the module.
                module.unfreeze()
            else:
                logging.debug("Module `{}` is not trainable so cannot be unfrozen".format(name))

    def save_to(self, filename: str, module_names: Optional[List[str]] = None):
        """
        Saves the state of trainable modules in the graph to a checkpoint file.

        Args:
            filename (string): Name of the file where the checkpoint will be saved.
            module_names: List of modules to be frozen (Optional). If passed, all modules will be saved.
        Raises:
            KeyError: If name of the module won't be recognized.
        """
        # Work on all modules.
        if module_names is None:
            module_names = self._modules.keys()

        # Prepare the "graph checkpoint".
        chkpt = {"header": {"nemo_core_version": nemo_version, "name": self.name}, "modules": {}}

        log_str = ''
        # Iterate through the modules one by one.
        for name in module_names:
            if name not in self._modules.keys():
                raise KeyError("Module `{}` not present in the `{}` graph".format(name, self.name))
            # Check module type.
            module = self._modules[name]
            if module.type == ModuleType.trainable:
                # Get module state_dict().
                chkpt["modules"][name] = get_state_dict(module)
                log_str += "  * Module '{}' ({}) params saved \n".format(module.name, type(module).__name__)
            else:
                logging.debug("Module `{}` is not trainable so cannot be saved".format(name))

        # Save checkpoint.
        save(chkpt, filename)
        log_str = "Saved  the '{}' graph to a checkpoint `{}`:\n".format(self.name, filename) + log_str
        logging.info(log_str)

    def restore_from(self, filename: str, module_names: Optional[List[str]] = None):
        """
        Restores the state of trainable modules in the graph from a checkpoint file.

        Args:
            filename (string): Name of the checkpoint to be restored from.
            module_names: List of modules to be frozen (Optional). If passed, all modules will be restored.
        Raises:
            KeyError: If name of the module won't be recognized.
        """
        # Work on all modules.
        if module_names is None:
            module_names = self._modules.keys()

        # Load the checkpoint.
        chkpt = load(filename)

        log_str = "Loading modules constituting the '{}' graph from the `{}` checkpoint :\n".format(
            chkpt["header"]["name"], filename
        )

        warning = False
        # Iterate through the modules one by one.
        for name in module_names:
            try:
                # Get module.
                module = self._modules[name]
                if module.type == ModuleType.trainable:
                    # Restore module weights
                    set_state_dict(module, chkpt["modules"][name])
                    log_str += "  * Module '{}' ({}) params loaded\n".format(module.name, type(module).__name__)
            except KeyError:
                log_str += "  ! Module '{}' params not found in checkpoint\n".format(name)
                warning = True

        # Log results.
        if warning:
            logging.warning(log_str)
        else:
            logging.info(log_str)

    @property
    def is_complete(self) -> bool:
        """
        Property checks if graph is "complete", which means that the graph has:
            * exactly one DataLayer
            * zero bound input ports

        Returns:
            True or false.
        """
        # We assume the first modules is DL.
        dl = self.modules[self.steps[0]]
        if dl.type != ModuleType.datalayer:
            return False

        # We assume there is only one DL in the whole graph.
        has_datalayer = False
        # Iterate through the modules one by one.
        for module in self._modules.values():
            # Get module type.
            if module.type == ModuleType.datalayer:
                if has_datalayer:
                    # More than one DL is not acceptable.
                    return False
                else:
                    has_datalayer = True

        # Now check the ports.
        if len(self._inputs) != 0:
            return False

        # Else:
        return True

    @property
    def is_trainable(self) -> bool:
        """
        Property checks if graph is "trainable", which means that the graph:
            * is in eval mode
            * has at least one Loss Neural Modules

        Returns:
            True or false.
        """
        # Check if gradients are collected in forward pass.
        if self.operation_mode == OperationMode.evaluation:
            return False

        # Iterate through the modules one by one.
        for module in self._modules.values():
            # Get module type.
            if module.type == ModuleType.loss:
                return True

        # No loss modules - cannot train the graph using back-propagation.
        return False

    def to(self, device_type: DeviceType, use_dataparallel: bool = False):
        """ 
        Moves the all the (trainable) modules belonging to a given graph to indicated device.

        """
        logging.info("Moving module(s) to {}".format(device_type))

        # Check device.
        if device_type == DeviceType.CPU:
            self._pt_device = torch.device("cpu")

        elif device_type == DeviceType.GPU:
            if not torch.cuda.is_available():
                raise ConfigurationError("Coudn't use GPUs as there is no CUDA installed")

            if torch.cuda.device_count() == 1:
                # Cannot use data parallel.
                if use_dataparallel:
                    raise ConfigurationError("Coudn't use Data Parallel as there is only one GPU device found")

            # Using cuda -- all devices.
            self._pt_device = torch.device("cuda")

        else:  # DeviceType.AllGpu : distributed data parallel.
            if not torch.cuda.is_available():
                raise ConfigurationError("Coudn't use Distributed Data Parallel as there is no CUDA installed")

            # Each worker will use a single gpu.
            use_dataparallel = False
            self._pt_device = torch.device("cuda")

        if use_dataparallel:
            logging.info("Using Data Parallelization on {} GPUs!".format(torch.cuda.device_count()))
            logging.warning("Data Parallel is currently under development, so use if it with caution")

        # Work on all modules.
        module_names = self._modules.keys()

        # Iterate through the modules one by one.
        for name in module_names:
            # Get module.
            module = self._modules[name]
            if module.type == ModuleType.trainable:
                # Check if we want to use data parallel.
                if use_dataparallel:
                    # Skip if the user explicitly said the module shouldn't be parallelized.
                    if not hasattr(module, 'skip_in_data_parallel') or not module.skip_in_data_parallel:
                        # Parallize model.
                        module = DataParallel(module)

                # Mode to device.
                self._modules[name] = module.to(self._pt_device)

    def configure_data_loader(
        self,
        batch_size=1,
        shuffle=False,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
        drop_last=False,
        timeout=0,
        worker_init_fn=None,
        multiprocessing_context=None,
    ):
        """
        Method updates the default parameters of the DataLoader used by a given neural graph.
        For the details on the function/meanings of the arguments, please refer to:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        """

        # Override all old values.
        self._data_loader_config = {
            "dataset": None,
            "batch_size": batch_size,
            "shuffle": shuffle,
            "sampler": sampler,
            "batch_sampler": batch_sampler,
            "num_workers": num_workers,
            "collate_fn": collate_fn,
            "pin_memory": pin_memory,
            "drop_last": drop_last,
            "timeout": timeout,
            "worker_init_fn": worker_init_fn,
            "multiprocessing_context": multiprocessing_context,
        }
        # If loader is set - reset it.
        self._data_loader = None

    def get_batch(self, yield_dict: bool = False) -> Union[Dict[str, Any], Tuple]:
        """
        Method yields a single batch. Optionally, instantiates the DataLoader object used by a given graph.
            
        Args:
            yield_dict: Flag used to yield a tuple or a dict - depending on the user needs
            (DEFAULT: False, i.e. yielding NamedTuples)
        Returns:
            Depending on the mode: a batch in the form of a singe namedtuple or of a dictionary.
        """
        # If loader is not set - set it.
        if self._data_loader is None:
            # Check graph.
            if not self.is_complete:
                raise ConfigurationError("Cannot get a batch as graph is incomplete!")
            # Get datalayer/dataset.
            dl = self.modules[self.steps[0]]
            # Make sure that this is DataLayer/dataset instance.
            if isinstance(dl, torch.utils.data.Dataset):
                self._data_loader_config["dataset"] = dl
            else:
                self._data_loader_config["dataset"] = dl.dataset

            self._data_loader = torch.utils.data.DataLoader(**self._data_loader_config)

        # Return tuple or dict - depending on the user needs.
        if yield_dict:
            # Fetch a batch - in the form of a dict.
            for batch in self._data_loader:
                if len(dl.output_ports.keys()) == 1:
                    # Dict with a single key:value pair.
                    batch = batch.to(self._pt_device) if isinstance(batch, torch.Tensor) else batch
                    yield {list(dl.output_ports.keys())[0]: batch}
                else:
                    # Yield a dictionary.
                    batch = [elem.to(self._pt_device) if isinstance(elem, torch.Tensor) else elem for elem in batch]
                    yield dict(zip(dl.output_ports.keys(), batch))

        else:
            # Create a tuple class - if required.
            if len(dl.output_ports.keys()) > 1:
                # Create a named tuple type enabling to access outputs by attributes (e.g. out.x).
                output_class_name = f'{dl.__class__.__name__}Output'
                result_type = namedtuple(typename=output_class_name, field_names=dl.output_ports.keys())

            # Fetch a batch - in the form of a tuple.
            for batch in self._data_loader:
                if len(dl.output_ports.keys()) == 1:
                    # Return a single object.
                    batch = batch.to(self._pt_device) if isinstance(batch, torch.Tensor) else batch
                    yield batch
                else:
                    # Return a tuple.
                    batch = [elem.to(self._pt_device) if isinstance(elem, torch.Tensor) else elem for elem in batch]
                    yield result_type(*batch)

    def __forward(self, inputs: Dict[str, Any], connections: Dict[str, Any]):
        """
        Graph forward method. In the execution it follows the order of the modules defined in steps and passes data
        between modules following the connectivity stored in NmTensors.
        """
        # Execute modules one by one.
        for step_number in range(len(self._steps)):
            # If graph is complete - we already fetched data from module 0 (DL)
            if self.is_complete and step_number == 0:
                # So lets skip it.
                continue

            # Get the module.
            module_name = self._steps[step_number]
            module = self._modules[module_name]

            # Change the module's operation mode.
            module.operation_mode = self.operation_mode

            # Produce list of arguments that will be passed to a given modules.
            module_args = {}
            # Do it by:
            # - harvesing input port names of a given module,
            # - checking if the input was not bound (in the inner graph),
            # - checking if we have already tensors leading to that input (in outer graph).
            for input_port_name in module.input_ports.keys():
                # Check if this port was bound in the inner graph.
                key = self.inputs.has_binding(step_number, input_port_name)
                # If so, then we must pass whatever was passed to that port in the list of arguments.
                if key is not None:
                    module_args[input_port_name] = inputs[key]
                    continue

                # Else: find a tensor that should be passed to the given module's input.
                producer_step = -1
                # Search for producer/port that we should use.
                for connection in connections:
                    if (
                        connection.consumer.step_number == step_number
                        and connection.consumer.module_name == module_name
                        and connection.consumer.port_name == input_port_name
                    ):
                        # Got the connection!
                        producer_step = connection.producer.step_number
                        # producer_name = connection.producer.module_name
                        producer_port_name = connection.producer.port_name
                        break
                # Make sure we found the producer.
                if producer_step == -1:
                    # Check if this port is not optional.
                    if not module.input_ports[input_port_name].optional:
                        err = "Couldn't find the producer of the {} input to the module {} called in step {}".format(
                            input_port_name, module_name, step_number
                        )
                        raise ValueError(err)
                    # If it is optional and not provided - simply do nothing...
                else:
                    # Add this to the argument
                    module_args[input_port_name] = self._forward_data[producer_step][producer_port_name]

            # Ok, now we have all keyword arguments. We can call() the module.
            module_outputs = module(force_pt=True, **module_args)

            # Collect all the produced output tensors and add them to this graph.
            output_names = list(module.output_ports.keys())

            if len(output_names) == 1:
                # Handle the case of a single data produced.
                self._forward_data[step_number][output_names[0]] = module_outputs
            else:
                # Compare module_outputs with the output port definitions.
                if len(output_names) != len(module_outputs):
                    err = "Invalid number of outputs produced by the module "
                    err += "{} - expected: `{}`, received: `{}`".format(module_name, output_names, len(module_outputs))
                    raise ValueError(err)

                # Add them to the passed data.
                for key, value in zip(output_names, module_outputs):
                    self._forward_data[step_number][key] = value

        # Ok, now return only the bound outputs.
        if len(self.outputs) == 1:
            # Return a single object.
            smp = self.outputs[list(self.outputs.keys())[0]].producer_step_module_port
            return self._forward_data[smp.step_number][smp.port_name]
        elif len(self.outputs) > 1:
            # Return a tuple.
            output_class_name = f'{self.__class__.__name__}Output'
            result_type = namedtuple(typename=output_class_name, field_names=self.outputs.keys())
            # Prepare list values.
            values = []
            for output in self.outputs.values():
                smp = output.producer_step_module_port
                values.append(self._forward_data[smp.step_number][smp.port_name])
            # Return the object
            return result_type(*values)
        else:  # (==0)
            # Generally this should not happen!
            return

    def forward(self, *args: Optional[Tuple], **kwargs: Optional[Dict[str, Any]]):
        """
        Graph forward method. In the execution it follows the order of the modules defined in steps and passes data
        between modules following the connectivity stored in NmTensors.

        Accepts a batch (passed as a tuple in args) OR a list of named arguments (as kwargs).

        Use-case 1:
        
        .. code-block:: python

            # Retrieve batch as a tuple and pass it to forward() as tuple.
            for batch in training_graph.get_batch():
                training_graph.forward(batch)

        Use-case 2:

        .. code-block:: python

            # Retrieve batch as a tuple and pass it to forward() as list of named arguments.
            for batch in training_graph.get_batch():
                training_graph.forward(input1=batch.input1, input2=batch.input2)


        Use-case 3:

        .. code-block:: python

            # Retrieve batch as dictionary and pass it to forward() as list of named arguments.
            for batch in training_graph.get_batch(yield_dict=True):
                training_graph.forward(**batch)

        Args:
            args: A tuple object (a batch) with all required inputs (optional)
            kwargs: a dictionary containing all required inputs (optional)

        Returns:
            A tuple object containing all graph outputs
        """

        # Get list of argument names.
        if self.is_complete:
            # Use dataset definitions.
            dl = self.modules[self.steps[0]]
            input_names = list(dl.output_ports.keys())
        else:
            input_names = list(self._inputs.keys())

        # Work on args or kwargs - depending on input_dict settings.
        if len(args) > 0:
            # Operate on args, i.e. a single tuple or a single "object" (1 input port to graph).
            if len(args) == 1:
                # Check if it is a tuple.
                if isinstance(args[0], tuple) and hasattr(args[0], "_fields"):
                    inputs = args[0]._asdict()
                else:
                    # There is only one input - use the input_names[0] as key.
                    inputs = {input_names[0]: args[0]}
            else:
                err = "Invalid argument passed to `graph forward()`"
                err += " - expected: a tuple, received: `{}`".format(args)
                raise ValueError(err)
        else:
            # Operate on named arguments.
            inputs = kwargs

        # Compare inputs with the desired inputs.
        if len(input_names) != len(inputs) or sorted(input_names) != sorted(inputs.keys()):
            err = "Invalid list of arguments passed to the `graph forward()`"
            err += " - expected: `{}`, received: `{}`".format(input_names, inputs.keys())
            raise ValueError(err)

        # Copy inputs to an adequate (module.output->value) structure.
        if self._forward_data is not None:
            self._forward_data = None
        self._forward_data = defaultdict(lambda: {})

        if self.is_complete:
            # Treat all the inputs as "DL outputs".
            for key, value in inputs.items():
                self._forward_data[0][key] = value

        # Create a list of connenctions: (producer.port -> consumer.port)
        connections = []
        for tensors in self.tensors.values():
            for t in tensors.values():
                connections.extend(t.connections())

        if self.operation_mode == OperationMode.evaluation:
            # Perform forward - without collecting of the gradients.
            with torch.no_grad():
                return self.__forward(inputs, connections)
        else:
            # Perform forward - and collect all the gradients.
            return self.__forward(inputs, connections)

    def backward(self, losses: List["Tensor"] = []):
        """
        Backward pass through the graph. Optionally the method accepts list of losses to backpropagate from.
        If not provided, will collect all output of all loss modules in the graph and backpropagate from them.

        Args:
            losses: list of losses to backpropagate from (OPTIONAL, DEFAULT: [])

        Raises:
            ConfigurationError if graph is non-trainable.
        """
        # Check if we can run backwards at all.
        if not self.is_trainable:
            raise ConfigurationError(
                "Cannot run backward propagation as the graph `{}`` is non-trainable".format(self.name)
            )

        losses_to_backpropagate = []
        # If losses are passed - use those during backpropagation.
        if len(losses) != 0:
            losses_to_backpropagate = losses
        else:
            # Else: collect outputs of all Loss NMs.
            for step_number in range(len(self._steps)):

                # Get the module.
                module_name = self._steps[step_number]
                module = self._modules[module_name]

                # Collect module outputs.
                # Assumption: loss modules return only loss.
                if module.type == ModuleType.loss:
                    for output_port_names in module.output_ports.keys():
                        losses_to_backpropagate.append(self._forward_data[step_number][output_port_names])

        # Estimate the total number of backward passes (one from each tensor).
        total_passes = len(losses_to_backpropagate)

        # All but the last call to backward should have the retain_graph=True option.
        pass_counter = 0
        for loss in losses_to_backpropagate:
            pass_counter += 1
            if pass_counter == total_passes:
                # Last pass.
                loss.backward()
            else:
                # "Other pass."
                loss.backward(retain_graph=True)

    def parameters(self, recurse: bool = True):
        """
        Args:
            recurse (bool): if True, then yields parameters of graph modules and all their submodules.
        Returns:
            An iterator over parameter of all trainable modules.
        """
        for module_name in self._modules:
            # Get module.
            module = self._modules[module_name]
            if module.type == ModuleType.trainable:
                # Yield it parameters.
                for _, param in module.named_parameters(recurse=recurse):
                    yield param

    def named_parameters(self, recurse: bool = True):
        """
        Args:
            recurse (bool): if True, then yields parameters of graph modules and all their submodules.
        Returns:
            An iterator ovar all named parameters of all trainable components.
        """
        for module_name in self._modules:
            # Get module.
            module = self._modules[module_name]
            if module.type == ModuleType.trainable:
                # Yield it parameters.
                for name, param in module.named_parameters(recurse=recurse):
                    yield name, param



# Copyright (c) 2019 NVIDIA Corporation
import librosa
import matplotlib.pylab as plt
import numpy as np
import torch

from nemo.utils import logging

__all__ = [
    "waveglow_log_to_tb_func",
    "waveglow_process_eval_batch",
    "waveglow_eval_log_to_tb_func",
    "tacotron2_log_to_tb_func",
    "tacotron2_process_eval_batch",
    "tacotron2_process_final_eval",
    "tacotron2_eval_log_to_tb_func",
]


def griffin_lim(magnitudes, n_iters=50, n_fft=1024):
    """
    Griffin-Lim algorithm to convert magnitude spectrograms to audio signals
    """
    phase = np.exp(2j * np.pi * np.random.rand(*magnitudes.shape))
    complex_spec = magnitudes * phase
    signal = librosa.istft(complex_spec)
    if not np.isfinite(signal).all():
        logging.warning("audio was not finite, skipping audio saving")
        return np.array([0])

    for _ in range(n_iters):
        _, phase = librosa.magphase(librosa.stft(signal, n_fft=n_fft))
        complex_spec = magnitudes * phase
        signal = librosa.istft(complex_spec)
    return signal


def waveglow_log_to_tb_func(
    swriter,
    tensors,
    step,
    tag="train",
    log_images=False,
    log_images_freq=1,
    n_fft=1024,
    hop_length=256,
    window="hann",
    mel_fb=None,
):
    loss, audio_pred, spec_target, mel_length = tensors
    if loss:
        swriter.add_scalar("loss", loss, step)
    if log_images and step % log_images_freq == 0:
        mel_length = mel_length[0]
        spec_target = spec_target[0].data.cpu().numpy()[:, :mel_length]
        swriter.add_image(
            f"{tag}_mel_target", plot_spectrogram_to_numpy(spec_target), step, dataformats="HWC",
        )
        if mel_fb is not None:
            mag, _ = librosa.core.magphase(
                librosa.core.stft(
                    np.nan_to_num(audio_pred[0].cpu().detach().numpy()),
                    n_fft=n_fft,
                    hop_length=hop_length,
                    window=window,
                )
            )
            mel_pred = np.matmul(mel_fb.cpu().numpy(), mag).squeeze()
            log_mel_pred = np.log(np.clip(mel_pred, a_min=1e-5, a_max=None))
            swriter.add_image(
                f"{tag}_mel_predicted",
                plot_spectrogram_to_numpy(log_mel_pred[:, :mel_length]),
                step,
                dataformats="HWC",
            )


def waveglow_process_eval_batch(tensors: dict, global_vars: dict):
    if 'tensorboard' not in global_vars.keys():
        global_vars['tensorboard'] = {}
        for k, v in tensors.items():
            if k.startswith("processed_signal"):
                global_vars['tensorboard']['mel_target'] = v[0]
            if k.startswith("audio"):
                global_vars['tensorboard']['audio_pred'] = v[0]
            if k.startswith("processed_length"):
                global_vars['tensorboard']['mel_length'] = v[0]


def waveglow_eval_log_to_tb_func(
    swriter, global_vars, step, tag=None, n_fft=1024, hop_length=256, window="hann", mel_fb=None,
):
    spec_target = global_vars['tensorboard']["mel_target"]
    audio_pred = global_vars['tensorboard']["audio_pred"]
    mel_length = global_vars['tensorboard']['mel_length']
    waveglow_log_to_tb_func(
        swriter,
        [None, audio_pred, spec_target, mel_length],
        step,
        tag=tag,
        log_images=True,
        n_fft=n_fft,
        hop_length=hop_length,
        window=window,
        mel_fb=mel_fb,
    )


def tacotron2_log_to_tb_func(
    swriter,
    tensors,
    step,
    tag="train",
    log_images=False,
    log_images_freq=1,
    add_audio=True,
    griffin_lim_mag_scale=1024,
    griffin_lim_power=1.2,
    sr=22050,
    n_fft=1024,
    n_mels=80,
    fmax=8000,
):
    loss, spec_target, mel_postnet, gate, gate_target, alignments = tensors
    if loss:
        swriter.add_scalar("loss", loss, step)
    if log_images and step % log_images_freq == 0:
        swriter.add_image(
            f"{tag}_alignment", plot_alignment_to_numpy(alignments[0].data.cpu().numpy().T), step, dataformats="HWC",
        )
        swriter.add_image(
            f"{tag}_mel_target", plot_spectrogram_to_numpy(spec_target[0].data.cpu().numpy()), step, dataformats="HWC",
        )
        swriter.add_image(
            f"{tag}_mel_predicted",
            plot_spectrogram_to_numpy(mel_postnet[0].data.cpu().numpy()),
            step,
            dataformats="HWC",
        )
        swriter.add_image(
            f"{tag}_gate",
            plot_gate_outputs_to_numpy(gate_target[0].data.cpu().numpy(), torch.sigmoid(gate[0]).data.cpu().numpy(),),
            step,
            dataformats="HWC",
        )
        if add_audio:
            filterbank = librosa.filters.mel(sr=sr, n_fft=n_fft, n_mels=n_mels, fmax=fmax)
            log_mel = mel_postnet[0].data.cpu().numpy().T
            mel = np.exp(log_mel)
            magnitude = np.dot(mel, filterbank) * griffin_lim_mag_scale
            audio = griffin_lim(magnitude.T ** griffin_lim_power)
            swriter.add_audio(f"audio/{tag}_predicted", audio / max(np.abs(audio)), step, sample_rate=sr)

            log_mel = spec_target[0].data.cpu().numpy().T
            mel = np.exp(log_mel)
            magnitude = np.dot(mel, filterbank) * griffin_lim_mag_scale
            audio = griffin_lim(magnitude.T ** griffin_lim_power)
            swriter.add_audio(f"audio/{tag}_target", audio / max(np.abs(audio)), step, sample_rate=sr)


def tacotron2_process_eval_batch(tensors: dict, global_vars: dict):
    if 'EvalLoss' not in global_vars.keys():
        global_vars['EvalLoss'] = []
    if 'tensorboard' not in global_vars.keys():
        global_vars['tensorboard'] = {}
        for k, v in tensors.items():
            if k.startswith("processed_signal"):
                global_vars['tensorboard']['mel_target'] = v[0]
            if k.startswith("mel_output"):
                global_vars['tensorboard']['mel_pred'] = v[0]
            if k.startswith("gate_output"):
                global_vars['tensorboard']['gate'] = v[0]
            if k.startswith("alignments"):
                global_vars['tensorboard']['alignments'] = v[0]
            if k.startswith("gate_target"):
                global_vars['tensorboard']['gate_target'] = v[0]

    for k in tensors.keys():
        if k.startswith("loss"):
            loss_key = k
    global_vars['EvalLoss'].append(torch.mean(torch.stack(tensors[loss_key])))


def tacotron2_process_final_eval(global_vars: dict, tag=None):
    eloss = torch.mean(torch.stack(global_vars['EvalLoss'])).item()
    global_vars['EvalLoss'] = eloss
    logging.info(f"==========>>>>>>Evaluation Loss {tag}: {eloss}")
    return global_vars


def tacotron2_eval_log_to_tb_func(
    swriter,
    global_vars,
    step,
    tag=None,
    add_audio=True,
    griffin_lim_mag_scale=1024,
    griffin_lim_power=1.2,
    sr=22050,
    n_fft=1024,
    n_mels=80,
    fmax=8000,
):
    spec_target = global_vars['tensorboard']["mel_target"]
    mel_postnet = global_vars['tensorboard']["mel_pred"]
    gate = global_vars['tensorboard']["gate"]
    gate_target = global_vars['tensorboard']["gate_target"]
    alignments = global_vars['tensorboard']["alignments"]
    swriter.add_scalar(f"{tag}.loss", global_vars['EvalLoss'], step)
    tacotron2_log_to_tb_func(
        swriter,
        [None, spec_target, mel_postnet, gate, gate_target, alignments],
        step,
        tag=tag,
        log_images=True,
        add_audio=add_audio,
        griffin_lim_mag_scale=griffin_lim_mag_scale,
        griffin_lim_power=griffin_lim_power,
        sr=sr,
        n_fft=n_fft,
        n_mels=n_mels,
        fmax=fmax,
    )


def save_figure_to_numpy(fig):
    # save it to a numpy array.
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    return data


def plot_alignment_to_numpy(alignment, info=None):
    fig, ax = plt.subplots(figsize=(6, 4))
    im = ax.imshow(alignment, aspect='auto', origin='lower', interpolation='none')
    fig.colorbar(im, ax=ax)
    xlabel = 'Decoder timestep'
    if info is not None:
        xlabel += '\n\n' + info
    plt.xlabel(xlabel)
    plt.ylabel('Encoder timestep')
    plt.tight_layout()

    fig.canvas.draw()
    data = save_figure_to_numpy(fig)
    plt.close()
    return data


def plot_spectrogram_to_numpy(spectrogram):
    fig, ax = plt.subplots(figsize=(12, 3))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower", interpolation='none')
    plt.colorbar(im, ax=ax)
    plt.xlabel("Frames")
    plt.ylabel("Channels")
    plt.tight_layout()

    fig.canvas.draw()
    data = save_figure_to_numpy(fig)
    plt.close()
    return data


def plot_gate_outputs_to_numpy(gate_targets, gate_outputs):
    fig, ax = plt.subplots(figsize=(12, 3))
    ax.scatter(
        range(len(gate_targets)), gate_targets, alpha=0.5, color='green', marker='+', s=1, label='target',
    )
    ax.scatter(
        range(len(gate_outputs)), gate_outputs, alpha=0.5, color='red', marker='.', s=1, label='predicted',
    )

    plt.xlabel("Frames (Green target, Red predicted)")
    plt.ylabel("Gate State")
    plt.tight_layout()

    fig.canvas.draw()
    data = save_figure_to_numpy(fig)
    plt.close()
    return data



# Copyright (c) 2019 NVIDIA Corporation
"""This file contains a collection of overly simplistic NeuralModules"""
import numpy as np
import torch as t
import torch.nn as nn
import torch.utils.data as t_utils

from nemo.backends.pytorch.nm import DataLayerNM, LossNM, TrainableNM
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs


class TaylorNet(TrainableNM):  # Note inheritance from TrainableNM
    """Module which learns Taylor's coefficients."""

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        Returns:
          A (dict) of module's input ports names to NeuralTypes mapping
        """
        return {"x": NeuralType(('B', 'D'), ChannelType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        Returns:
          A (dict) of module's output ports names to NeuralTypes mapping
        """
        return {"y_pred": NeuralType(('B', 'D'), ChannelType())}

    def __init__(self, dim, name=None):
        """
            Creates TaylorNet object.

            Args:
                dim: Number of dimensions (number of terms in Taylor series).
                name: Name of the module instance
        """
        super().__init__(name=name)

        # And of Neural Modules specific part. Rest is Pytorch code
        self._dim = dim
        self.fc1 = nn.Linear(self._dim, 1)
        t.nn.init.xavier_uniform_(self.fc1.weight)
        self.to(self._device)

    # IMPORTANT: input arguments to forward must match input input ports' names
    def forward(self, x):
        lst = []
        for pw in range(self._dim):
            lst.append(x ** pw)
        nx = t.cat(lst, dim=-1)
        return self.fc1(nx)


class RealFunctionDataLayer(DataLayerNM):
    """
    Data layer that yields (x, f(x)) data and label pairs.

    Args:
        n: Total number of samples
        batch_size: Size of each batch per iteration
        f_name: Name of the function that will be applied to each x value to get labels.
           Must take a torch tensor as input, and output a torch tensor of
           the same shape. Defaults to torch.sin().
           [Options: sin | cos]
        x_lo: Lower bound of domain to sample
        x_hi: Upper bound of domain to sample
    """

    def __len__(self):
        return self._n

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports
        """
        return {
            "x": NeuralType(('B', 'D'), ChannelType()),
            "y": NeuralType(('B', 'D'), LabelsType()),
        }

    def __init__(self, batch_size, f_name="sin", n=1000, x_lo=-4, x_hi=4, name=None):
        """
            Creates a datalayer returning (x-y) pairs, with n points from a given range.

            Args:
                batch_size: size of batch
                f_name: name of function ["sin" | "cos"]
                n: number of points
                x_lo: lower boundary along x axis
                x_hi: higher boundary along x axis
                name: Name of the module instance
        """
        super().__init__(name=name)

        # Dicionary with handled functions.
        handled_funcs = {"sin": t.sin, "cos": t.cos}

        # Get function - raises an exception if function is not handled
        func = handled_funcs[f_name]

        self._n = n
        self._batch_size = batch_size

        x_data = t.tensor(np.random.uniform(low=x_lo, high=x_hi, size=self._n)).unsqueeze(-1)
        y_data = func(x_data)
        self._dataset = t_utils.TensorDataset(x_data.float(), y_data.float())
        self._data_iterator = t_utils.DataLoader(self._dataset, batch_size=self._batch_size,)

    @property
    def data_iterator(self):
        return self._data_iterator

    @property
    def dataset(self):
        return self._dataset


class MSELoss(LossNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        predictions:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag)

        target:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag)
        """
        return {
            "predictions": NeuralType(('B', 'D'), ChannelType()),
            "target": NeuralType(('B', 'D'), LabelsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, name=None):
        super().__init__(name=name)
        self._criterion = nn.MSELoss()

    def _loss_function(self, **kwargs):
        return self._criterion(*(kwargs.values()))


class L1Loss(LossNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "predictions": NeuralType(('B', 'D'), ChannelType()),
            "target": NeuralType(('B', 'D'), LabelsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, name=None):
        super().__init__(name=name)
        self._criterion = nn.L1Loss()

    def _loss_function(self, **kwargs):
        return self._criterion(*(kwargs.values()))


class CrossEntropyLoss(LossNM):
    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "predictions": NeuralType(('B', 'D'), ChannelType()),
            "labels": NeuralType(tuple('B'), LabelsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, name=None):
        super().__init__(name=name)
        self._criterion = nn.CrossEntropyLoss()

    # You need to implement this function
    def _loss_function(self, **kwargs):
        return self._criterion(*(kwargs.values()))



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


from collections import namedtuple

# Tuple used for storing "step number", "module name" and "port name".
# (used in NmTensor's producer/consumer, port binding etc.).
# Module name is redundant, as it can be recovered from the step number.
StepModulePort = namedtuple('StepModulePort', ["step_number", "module_name", "port_name"])


# Tuple used for connection between a single producer and a single consummer consumer.
# (used in NmTensor's producer/consumer, port binding etc.).
Connection = namedtuple('Connection', ["producer", "consumer", "ntype"])



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Pytorch Dataset for training Neural Machine Translation."""

from collections import OrderedDict

import numpy as np
from torch.utils.data import Dataset

from nemo.collections.nlp.data.datasets.datasets_utils.data_preprocessing import dataset_to_ids

__all__ = ['TranslationDataset']


class TranslationDataset(Dataset):
    def __init__(self, tokenizer_src, tokenizer_tgt, dataset_src, dataset_tgt, tokens_in_batch=1024, clean=False):

        self.src_tokenizer = tokenizer_src
        self.tgt_tokenizer = tokenizer_tgt
        self.tokens_in_batch = tokens_in_batch

        src_ids = dataset_to_ids(dataset_src, tokenizer_src)
        tgt_ids = dataset_to_ids(dataset_tgt, tokenizer_tgt)
        if clean:
            src_ids, tgt_ids = self.clean_src_and_target(src_ids, tgt_ids)
        self.batch_indices = self.pack_data_into_batches(src_ids, tgt_ids)
        self.batches = self.pad_batches(src_ids, tgt_ids, self.batch_indices)

    def __len__(self):
        return len(self.batches)

    def __getitem__(self, idx):
        src_ids = self.batches[idx]["src"]
        tgt = self.batches[idx]["tgt"]
        labels = tgt[:, 1:]
        tgt_ids = tgt[:, :-1]
        src_mask = (src_ids != self.src_tokenizer.pad_id).astype(np.int32)
        tgt_mask = (tgt_ids != self.tgt_tokenizer.pad_id).astype(np.int32)
        sent_ids = self.batch_indices[idx]
        return src_ids, src_mask, tgt_ids, tgt_mask, labels, sent_ids

    def pad_batches(self, src_ids, tgt_ids, batch_indices):
        """
        Augments source and target ids in the batches with padding symbol
        to make the lengths of all sentences in the batches equal.
        """

        batches = {}
        for batch_idx, b in enumerate(batch_indices):
            src_len = max([len(src_ids[i]) for i in b])
            tgt_len = max([len(tgt_ids[i]) for i in b])
            src_ids_ = self.src_tokenizer.pad_id * np.ones((len(b), src_len), dtype=np.int)
            tgt_ids_ = self.tgt_tokenizer.pad_id * np.ones((len(b), tgt_len), dtype=np.int)
            for i, sentence_idx in enumerate(b):
                src_ids_[i][: len(src_ids[sentence_idx])] = src_ids[sentence_idx]
                tgt_ids_[i][: len(tgt_ids[sentence_idx])] = tgt_ids[sentence_idx]
            batches[batch_idx] = {"src": src_ids_, "tgt": tgt_ids_}
        return batches

    def pack_data_into_batches(self, src_ids, tgt_ids):
        """
        Takes two lists of source and target sentences, sorts them, and packs
        into batches to minimize the use of padding tokens. Returns a list of
        batches where each batch contains indices of sentences included into it
        """

        # create buckets sorted by the number of src tokens
        # each bucket is also sorted by the number of tgt tokens
        buckets = {}
        for i, src_id in enumerate(src_ids):
            src_len, tgt_len = len(src_id), len(tgt_ids[i])
            if src_len not in buckets.keys():
                buckets[src_len] = [(tgt_len, i)]
            else:
                buckets[src_len].append((tgt_len, i))

        for b_idx in buckets:
            buckets[b_idx] = sorted(buckets[b_idx])

        buckets = OrderedDict(sorted(buckets.items()))
        indices = list(buckets.keys())

        batches = [[]]
        num_batches = 0
        batch_size = 0
        i = 0
        src_len = 0
        tgt_len = 0

        while i < len(buckets.keys()):

            while buckets[indices[i]]:

                i_src = max(src_len, indices[i])
                i_tgt = max(tgt_len, buckets[indices[i]][0][0])

                try:
                    ip1_src = max(src_len, indices[i + 1])
                    ip1_tgt = max(tgt_len, buckets[indices[i + 1]][0][0])
                except IndexError:
                    ip1_src = i_src + 1
                    ip1_tgt = i_tgt + 1

                if i_src + i_tgt <= ip1_src + ip1_tgt:
                    src_len = i_src
                    tgt_len = i_tgt
                    _, idx = buckets[indices[i]].pop(0)
                else:
                    src_len = ip1_src
                    tgt_len = ip1_tgt
                    _, idx = buckets[indices[i + 1]].pop(0)

                batches[num_batches].append(idx)
                batch_size += 1

                if batch_size * (src_len + tgt_len) > self.tokens_in_batch:

                    num_examples_to_split = len(batches[num_batches])
                    batches_to_evict = 8 * ((num_examples_to_split - 1) // 8)

                    if batches_to_evict == 0:
                        batches_to_evict = num_examples_to_split

                    batches.append(batches[num_batches][batches_to_evict:])
                    batches[num_batches] = batches[num_batches][:batches_to_evict]
                    batch_size = num_examples_to_split - batches_to_evict

                    num_batches += 1
                    if batch_size > 0:
                        src_len = max([len(src_ids[j]) for j in batches[num_batches]])
                        tgt_len = max([len(tgt_ids[j]) for j in batches[num_batches]])
                    else:
                        src_len = 0
                        tgt_len = 0
                    break

            if not buckets[indices[i]]:
                i = i + 1

        if not batches[-1]:
            batches.pop(-1)

        return batches

    def clean_src_and_target(
        self, src_ids, tgt_ids, max_tokens=128, min_tokens=3, max_tokens_diff=25, max_tokens_ratio=2.5
    ):
        """
        Cleans source and target sentences to get rid of noisy data.
        Specifically, a pair of sentences is removed if
          -- either source or target is longer than *max_tokens*
          -- either source or target is shorter than *min_tokens*
          -- absolute difference between source and target is larger than
             *max_tokens_diff*
          -- one sentence is *max_tokens_ratio* times longer than the other
        """

        if len(src_ids) != len(tgt_ids):
            raise ValueError("Source and target corpora have different lengths!")
        src_ids_, tgt_ids_ = [], []
        for i in range(len(src_ids)):
            src_len, tgt_len = len(src_ids[i]), len(tgt_ids[i])
            if (
                src_len > max_tokens
                or tgt_len > max_tokens
                or src_len < min_tokens
                or tgt_len < min_tokens
                or (src_ids[i] == tgt_ids[i])
                or np.abs(src_len - tgt_len) > max_tokens_diff
            ):
                continue
            ratio = max(src_len - 2, 1) / max(tgt_len - 2, 1)
            if ratio > max_tokens_ratio or ratio < (1 / max_tokens_ratio):
                continue
            src_ids_.append(src_ids[i])
            tgt_ids_.append(tgt_ids[i])
        return src_ids_, tgt_ids_



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.backends.pytorch import DataLayerNM
from nemo.collections.nlp.data.datasets.sgd_dataset.sgd_dataset import SGDDataset
from nemo.core.neural_types import ChannelType, LabelsType, LengthsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['SGDDataLayer']


class SGDDataLayer(DataLayerNM):
    """
    Data layer for Schema Guided Dialogue State Tracking Dataset.
    Args:
        dataset_split (str): train/ dev/ test,
        dialogues_processor (obj):  containt dialogue data,
        dataset_type (Dataset): Dataset Type,
        shuffle (bool): enables shuffling, default=False
        num_workers (int): number of workers
        batch_size (int): batch size
        pin_memory (bool): enables copying Tensors into CUDA pinned memory before returning them
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        example_id_num (int): example ids
        service_id  (int): service ids
        is_real_example (bool): flag to determine is the example is valid
        utterance_ids (int): utterance ids
        utterance_segment (int): Denotes the identity of the sequence. Takes values 0 (system utterance) and 1 (user utterance)
        utterance_mask (int): Mask which takes the value 0 for padded tokens and 1 otherwise
        categorical_slot_status (int): The status of each categorical slot in the service
        cat_slot_status_mask(int): Masks out categorical status for padded cat slots, takes values 0 and 1
        categorical_slot_values (int): The index of the correct value for each categorical slot
        cat_slot_values_mask (int): Masks out categorical slots values for slots not used in the service, takes values 0 and 1
        noncategorical_slot_status (int): The status of each non-categorical slot in the service
        noncat_slot_status_mask(int): Masks out non-categorical status for padded cat slots, takes values 0 and 1
        noncategorical_slot_value_start (int): The index of the starting subword corresponding to the slot span for a non-categorical slot value
        noncategorical_slot_value_end (int): The index of the ending (inclusive) subword corresponding to the slot span for a non-categorical slot value
        start_char_idx (int): Start character indices in the original utterance corresponding to the tokens
        end_char_idx (int): Inclusive end character indices in the original utterance corresponding to the tokens
        num_slots (int): Total number of slots present in the service
        requested_slot_status (int): Takes value 1 if the corresponding slot is requested, 0 otherwise
        req_slot_mask (int): Masks requested slots not used for the particular service
        intent_status_mask (long): Masks out padded intents in the service, takes values 0 and 1
        intent_status_labels (int): Intent labels

        """
        return {
            "example_id_num": NeuralType(('B'), ChannelType()),
            "service_id": NeuralType(('B'), ChannelType()),
            "is_real_example": NeuralType(('B'), ChannelType()),
            "utterance_ids": NeuralType(('B', 'T'), ChannelType()),
            "utterance_segment": NeuralType(('B', 'T'), ChannelType()),
            "utterance_mask": NeuralType(('B', 'T'), ChannelType()),
            "categorical_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "cat_slot_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "categorical_slot_values": NeuralType(('B', 'T'), LabelsType()),
            "cat_slot_values_mask": NeuralType(('B', 'T', 'C'), ChannelType()),
            "noncategorical_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "noncat_slot_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "noncategorical_slot_value_start": NeuralType(('B', 'T'), LabelsType()),
            "noncategorical_slot_value_end": NeuralType(('B', 'T'), LabelsType()),
            "start_char_idx": NeuralType(('B', 'T'), LabelsType()),
            "end_char_idx": NeuralType(('B', 'T'), LabelsType()),
            "num_slots": NeuralType(('B'), LengthsType()),
            "requested_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "req_slot_mask": NeuralType(('B', 'T'), ChannelType()),
            "intent_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "intent_status_labels": NeuralType(('B'), LabelsType()),
        }

    def __init__(
        self,
        dataset_split,
        dialogues_processor,
        dataset_type=SGDDataset,
        shuffle=False,
        batch_size=1,
        num_workers=-1,
        pin_memory=False,
    ):
        super().__init__()
        dataset_params = {
            'dataset_split': dataset_split,
            'dialogues_processor': dialogues_processor,
        }
        self._dataset = dataset_type(**dataset_params)
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._pin_memory = pin_memory
        if num_workers >= 0:
            self._num_workers = num_workers

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Prediction and evaluation-related utility functions.
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/pred_utils.py
"""

import json
import os
from collections import OrderedDict, defaultdict

from nemo import logging
from nemo.collections.nlp.data.datasets.sgd_dataset.input_example import (
    STATUS_ACTIVE,
    STATUS_CARRY,
    STATUS_DONTCARE,
    STATUS_OFF,
    STR_DONTCARE,
)

REQ_SLOT_THRESHOLD = 0.5

# MIN_SLOT_RELATION specifes the minimum number of relations between two slots in the training dialogues to get considered for carry-over
MIN_SLOT_RELATION = 0.1

__all__ = ['get_predicted_dialog_baseline', 'write_predictions_to_file']


def carry_over_slots(
    cur_usr_frame,
    all_slot_values,
    slots_relation_list,
    frame_service_prev,
    slot_values,
    sys_slots_agg,
    sys_slots_last,
):
    """This function searches the candidate list for cross-service cases to find and update the values for all the slots in the current predicted state
    It is called when state is predicted for a frame.
    Args:
        cur_usr_frame: the current frame of the user
        all_slot_values: dictionary of all the slots and their values extracted from the dialogue until the current turn for all services
        slots_relation_list: list of the candidates for carry-over for each (service, slot)
        frame_service_prev: the service of the last system's frame
        sys_slots_last: dictionary of all the slots and values mentioned in the last system utterance
        sys_slots_agg:  dictionary of all the slots and values mentioned in the all the system utterances until the current turn
        slot_values: dictionary of all the slots and their values extracted from the dialogue until the current turn for the current service
      Returns:
        the extracted value for the slot
    """

    if frame_service_prev == "" or frame_service_prev == cur_usr_frame["service"]:
        return
    for (service_dest, slot_dest), cands_list in slots_relation_list.items():
        if service_dest != cur_usr_frame["service"]:
            continue
        for service_src, slot_src, freq in cands_list:
            if freq < MIN_SLOT_RELATION or service_src != frame_service_prev:
                continue
            if service_src in all_slot_values and slot_src in all_slot_values[service_src]:
                slot_values[slot_dest] = all_slot_values[service_src][slot_src]
            if service_src in sys_slots_agg and slot_src in sys_slots_agg[service_src]:
                slot_values[slot_dest] = sys_slots_agg[service_src][slot_src]
            if service_src in sys_slots_last and slot_src in sys_slots_last[service_src]:
                slot_values[slot_dest] = sys_slots_last[service_src][slot_src]


def get_carryover_value(
    slot,
    cur_usr_frame,
    all_slot_values,
    slots_relation_list,
    frame_service_prev,
    sys_slots_last,
    sys_slots_agg,
    sys_rets,
):
    """This function searches the previous system actions and also the candidate list for cross-service cases to find a value for a slot
    It is called when a value for a slot can not be found the last user utterance
    Args:
        slot: name of the slot to find and extract a value for
        cur_usr_frame: the current frame of the user
        frame_service_prev: the service of the last system's frame
        all_slot_values: dictionary of all the slots and their values extracted from the dialogue until the current turn for all services
        sys_slots_last: dictionary of all the slots and values mentioned in the last system utterance
        sys_slots_agg:  dictionary of all the slots and values mentioned in the all the system utterances until the current turn
        slots_relation_list: list of the candidates for carry-over for each (service, slot)
        sys_rets: list of the extracted slots and values from system utterances until the current turn, used for debugging
      Returns:
        the extracted value for the slot
    """
    extracted_value = None
    if slot in sys_slots_agg[cur_usr_frame["service"]]:
        extracted_value = sys_slots_agg[cur_usr_frame["service"]][slot]
        sys_rets[slot] = extracted_value
    elif (cur_usr_frame["service"], slot) in slots_relation_list:
        cands_list = slots_relation_list[(cur_usr_frame["service"], slot)]
        for dmn, slt, freq in cands_list:
            if freq < MIN_SLOT_RELATION:
                continue
            if dmn in all_slot_values and slt in all_slot_values[dmn]:
                extracted_value = all_slot_values[dmn][slt]
            if dmn in sys_slots_agg and slt in sys_slots_agg[dmn]:
                extracted_value = sys_slots_agg[dmn][slt]
            if dmn in sys_slots_last and slt in sys_slots_last[dmn]:
                extracted_value = sys_slots_last[dmn][slt]
    return extracted_value


def get_predicted_dialog_nemotracker(dialog, all_predictions, schemas, eval_debug, in_domain_services):
    """This is NeMo Tracker which would be enabled by passing "--tracker_model=nemotracker".
    It improves the performance significantly by employing carry-over mechanism for in-service and cross-service.

    * **In-service carry-over mechanism**: There are cases that the value for some slots are not mentioned in the last user utterrance, but in the previous system utterances or actions.\
    Therefore, whenever the status of a non-categorical slot is active but no value can be found in the user utterance, we search the list of slots and their values mentioned in the previous system actions to find a value for this slot.
    The most recent value is used as the value for the slot. It is called in-domain carry-over as it happens inside a service.

    * **Cross-service carry-over mechanism**: In multi-domain dialogues, switching between two services can happen in the dialogue. In such cases, there can be some values to get transfered to the new service automatically.
    For instance when user is reserving flight tickets for two persons, it can be assumed that number of people for hotel reservation should also be two. To handle such cases, when we process the dialogues, we also record the list of these carry-over between two services from the training data.
    A candidate list for each (service, slot) is produced which show the list possible carry-over for that slot. These lists are stored in a file along with the processed dialogues and would be read and used in the state tracker to carry values when switches happens from one service to another.
    Whenever we find a switch and have an active non-categorical slot without any value, we would try to use that candidate list to retrieve a value for that slot from other slots in other services in previous turns. The latest value is used if multiple values are found.

    Args:
        dialog: A json object containing dialogue whose labels are to be updated.
        all_predictions: A dict mapping prediction name to the predicted value.
        schemas: A Schema object wrapping all the schemas for the dataset.
        eval_debug: specifies if it is running in DEBUG mode, so to generate the error analysis outputs
        in_domain_services: list of the seen services
    Returns:
        A json object containing the dialogue with labels predicted by the model.
  """

    dialog_id = dialog["dialogue_id"]
    # The slot values tracked for each service.
    all_slot_values = defaultdict(OrderedDict)
    sys_slots_agg = defaultdict(OrderedDict)
    sys_slots_last = defaultdict(OrderedDict)

    sys_rets = OrderedDict()
    true_state_prev = OrderedDict()
    true_state = OrderedDict()
    frame_service_prev = ""
    slots_relation_list = schemas._slots_relation_list

    for turn_idx, turn in enumerate(dialog["turns"]):
        if turn["speaker"] == "SYSTEM":
            sys_slots_last = defaultdict(OrderedDict)
            for frame in turn["frames"]:
                if frame["service"] not in sys_slots_agg:
                    sys_slots_agg[frame["service"]] = OrderedDict()
                if frame["service"] not in sys_slots_last:
                    sys_slots_last[frame["service"]] = OrderedDict()
                for action in frame["actions"]:
                    if action["slot"] and len(action["values"]) > 0:
                        sys_slots_agg[frame["service"]][action["slot"]] = action["values"][0]
                        sys_slots_last[frame["service"]][action["slot"]] = action["values"][0]
        elif turn["speaker"] == "USER":
            user_utterance = turn["utterance"]
            system_utterance = dialog["turns"][turn_idx - 1]["utterance"] if turn_idx else ""
            turn_id = "{:02d}".format(turn_idx)
            if len(turn["frames"]) == 2:
                if frame_service_prev != "" and turn["frames"][0]["service"] != frame_service_prev:
                    frame_tmp = turn["frames"][0]
                    turn["frames"][0] = turn["frames"][1]
                    turn["frames"][1] = frame_tmp

            for frame in turn["frames"]:
                cat_slot_status_acc = 0
                cat_slot_status_num = 0
                noncat_slot_status_num = 0
                noncat_slot_status_acc = 0
                cat_slot_value_acc = 0
                cat_slot_value_num = 0
                noncat_slot_value_acc = 0
                noncat_slot_value_num = 0

                predictions = all_predictions[(dialog_id, turn_id, frame["service"])]
                slot_values = all_slot_values[frame["service"]]
                service_schema = schemas.get_service_schema(frame["service"])

                predictions["cat_slot_status_p"] = predictions["cat_slot_status_p"].cpu().numpy()
                predictions["cat_slot_status"] = predictions["cat_slot_status"].cpu().numpy()
                predictions["cat_slot_value"] = predictions["cat_slot_value"].cpu().numpy()
                predictions["cat_slot_value_p"] = predictions["cat_slot_value_p"].cpu().numpy()

                predictions["noncat_slot_status_p"] = predictions["noncat_slot_status_p"].cpu().numpy()
                predictions["noncat_slot_status"] = predictions["noncat_slot_status"].cpu().numpy()
                predictions["noncat_slot_p"] = predictions["noncat_slot_p"].cpu().numpy()

                predictions["noncat_alignment_start"] = predictions["noncat_alignment_start"].cpu().numpy()
                predictions["noncat_alignment_end"] = predictions["noncat_alignment_end"].cpu().numpy()
                predictions["cat_slot_status_GT"] = predictions["cat_slot_status_GT"].cpu().numpy()
                predictions["noncat_slot_status_GT"] = predictions["noncat_slot_status_GT"].cpu().numpy()

                # Remove the slot spans and state if present.
                true_state_prev = [] if len(true_state) == 0 else true_state["slot_values"]
                true_slots = frame.pop("slots", None)
                true_state = frame.pop("state", None)

                # The baseline model doesn't predict slot spans. Only state predictions
                # are added.
                state = OrderedDict()

                # Add prediction for active intent. Offset is subtracted to account for
                # NONE intent.
                active_intent_id = predictions["intent_status"]
                state["active_intent"] = (
                    service_schema.get_intent_from_id(active_intent_id - 1) if active_intent_id else "NONE"
                )

                # Add prediction for requested slots.
                requested_slots = []
                for slot_idx, slot in enumerate(service_schema.slots):
                    if predictions["req_slot_status"][slot_idx] > REQ_SLOT_THRESHOLD:
                        requested_slots.append(slot)
                state["requested_slots"] = requested_slots

                # Add prediction for user goal (slot values).
                # Categorical slots.
                categorical_slots_dict = OrderedDict()
                non_categorical_slots_dict = OrderedDict()
                for slot_idx, slot in enumerate(service_schema.categorical_slots):
                    cat_slot_status_num += 1

                    slot_status = predictions["cat_slot_status"][slot_idx]
                    extracted_value = None
                    if slot_status == STATUS_DONTCARE:
                        extracted_value = STR_DONTCARE
                    elif slot_status == STATUS_ACTIVE:
                        carryover_value = get_carryover_value(
                            slot,
                            frame,
                            all_slot_values,
                            slots_relation_list,
                            frame_service_prev,
                            sys_slots_last,
                            sys_slots_agg,
                            sys_rets,
                        )

                        if (
                            service_schema.get_categorical_slot_values(slot)[predictions["cat_slot_value"][slot_idx]]
                            != "#CARRYVALUE#"
                            or carryover_value is None
                        ):
                            value_idx = predictions["cat_slot_value"][slot_idx]
                            extracted_value = service_schema.get_categorical_slot_values(slot)[value_idx]
                        else:
                            extracted_value = carryover_value
                    elif slot_status == STATUS_CARRY:
                        carryover_value = get_carryover_value(
                            slot,
                            frame,
                            all_slot_values,
                            slots_relation_list,
                            frame_service_prev,
                            sys_slots_last,
                            sys_slots_agg,
                            sys_rets,
                        )

                        if carryover_value is None:
                            extracted_value = None
                        else:
                            extracted_value = carryover_value

                    elif slot_status == STATUS_OFF:
                        extracted_value = None

                    if extracted_value is not None:
                        slot_values[slot] = extracted_value

                    # debugging info processing
                    if predictions["cat_slot_status_GT"][slot_idx] != predictions["cat_slot_status"][slot_idx] or (
                        predictions["cat_slot_status_GT"][slot_idx] == predictions["cat_slot_status"][slot_idx]
                        and predictions["cat_slot_status_GT"][slot_idx] != STATUS_OFF
                        and extracted_value not in true_state['slot_values'][slot]
                    ):
                        categorical_slots_dict[slot] = (
                            predictions["cat_slot_status_GT"][slot_idx],
                            predictions["cat_slot_status"][slot_idx],
                            predictions["cat_slot_status_p"][slot_idx],
                            service_schema.get_categorical_slot_values(slot)[predictions["cat_slot_value"][slot_idx]],
                            service_schema.get_categorical_slot_values(slot)[
                                predictions["cat_slot_value_GT"][slot_idx]
                            ],
                            extracted_value,
                            predictions["cat_slot_value_p"][slot_idx],
                        )

                    if predictions["cat_slot_status_GT"][slot_idx] == predictions["cat_slot_status"][slot_idx]:
                        cat_slot_status_acc += 1
                    if predictions["cat_slot_status_GT"][slot_idx] != STATUS_OFF:
                        cat_slot_value_num += 1
                        if extracted_value in true_state['slot_values'][slot]:
                            cat_slot_value_acc += 1
                    # debugging info processing ended

                for slot_idx, slot in enumerate(service_schema.non_categorical_slots):
                    noncat_slot_status_num += 1

                    tok_start_idx = predictions["noncat_slot_start"][slot_idx]
                    tok_end_idx = predictions["noncat_slot_end"][slot_idx]
                    ch_start_idx = predictions["noncat_alignment_start"][tok_start_idx]
                    ch_end_idx = predictions["noncat_alignment_end"][tok_end_idx]
                    extracted_value = None

                    slot_status = predictions["noncat_slot_status"][slot_idx]
                    if slot_status == STATUS_DONTCARE:
                        slot_values[slot] = STR_DONTCARE
                    elif slot_status == STATUS_ACTIVE:
                        if ch_start_idx > 0 and ch_end_idx > 0:
                            # Add span from the user utterance.
                            extracted_value = user_utterance[ch_start_idx - 1 : ch_end_idx]
                        else:
                            carryover_value = get_carryover_value(
                                slot,
                                frame,
                                all_slot_values,
                                slots_relation_list,
                                frame_service_prev,
                                sys_slots_last,
                                sys_slots_agg,
                                sys_rets,
                            )
                            extracted_value = carryover_value
                    elif slot_status == STATUS_CARRY:
                        carryover_value = get_carryover_value(
                            slot,
                            frame,
                            all_slot_values,
                            slots_relation_list,
                            frame_service_prev,
                            sys_slots_last,
                            sys_slots_agg,
                            sys_rets,
                        )

                        if carryover_value is None:
                            extracted_value = None
                        else:
                            extracted_value = carryover_value

                    if extracted_value is not None:
                        slot_values[slot] = extracted_value

                    # debugging info processing ended
                    if predictions["noncat_slot_status_GT"][slot_idx] != predictions["noncat_slot_status"][
                        slot_idx
                    ] or (
                        predictions["noncat_slot_status_GT"][slot_idx] == predictions["noncat_slot_status"][slot_idx]
                        and predictions["noncat_slot_status_GT"][slot_idx] != STATUS_OFF
                        and extracted_value not in true_state['slot_values'][slot]
                    ):
                        non_categorical_slots_dict[slot] = (
                            predictions["noncat_slot_status_GT"][slot_idx],
                            predictions["noncat_slot_status"][slot_idx],
                            predictions["noncat_slot_status_p"][slot_idx],
                            (ch_start_idx, ch_end_idx),
                            user_utterance[ch_start_idx - 1 : ch_end_idx]
                            if (ch_start_idx > 0 and ch_end_idx > 0)
                            else system_utterance[-ch_start_idx - 1 : -ch_end_idx],
                            extracted_value,
                            predictions["noncat_slot_p"][slot_idx],
                        )

                    if predictions["noncat_slot_status_GT"][slot_idx] != STATUS_OFF:
                        noncat_slot_value_num += 1
                        if extracted_value is not None and extracted_value in true_state['slot_values'][slot]:
                            noncat_slot_value_acc += 1

                    if predictions["noncat_slot_status_GT"][slot_idx] == predictions["noncat_slot_status"][slot_idx]:
                        noncat_slot_status_acc += 1
                    # debugging info processing ended

                carry_over_slots(
                    frame,
                    all_slot_values,
                    slots_relation_list,
                    frame_service_prev,
                    slot_values,
                    sys_slots_agg,
                    sys_slots_last,
                )
                # in debug mode, the following outputs would get generated which can be used for performing error analysis.
                # It prints out the information about the frames in the evaluation set which contain errors and those error in the predictaed state are not originated from previous frames or turns.
                # Therefore, these frames would be the origin of errors in the evaluation dialogues.
                # It just prints out the frames for seen services as our model is designed mostly for seen services and it does not work great on unseen ones.
                if eval_debug and frame["service"] in in_domain_services:
                    equal_state = True
                    for s, v in true_state['slot_values'].items():
                        if s not in slot_values or slot_values[s] not in v:
                            equal_state = False
                            break
                    for s, v in slot_values.items():
                        if s not in true_state['slot_values'] or v not in true_state['slot_values'][s]:
                            equal_state = False
                            break
                    if not equal_state:
                        cat_slot_status_acc = (
                            "NAN" if cat_slot_status_num == 0 else cat_slot_status_acc / cat_slot_status_num
                        )
                        noncat_slot_status_acc = (
                            "NAN" if noncat_slot_status_num == 0 else noncat_slot_status_acc / noncat_slot_status_num
                        )
                        cat_slot_value_acc = (
                            "NAN" if cat_slot_value_num == 0 else cat_slot_value_acc / cat_slot_value_num
                        )
                        noncat_slot_value_acc = (
                            "NAN" if noncat_slot_value_num == 0 else noncat_slot_value_acc / noncat_slot_value_num
                        )

                        found_err = False
                        if cat_slot_status_acc != "NAN" and cat_slot_status_acc < 1.0:
                            found_err = True
                        if noncat_slot_status_acc != "NAN" and noncat_slot_status_acc < 1.0:
                            found_err = True
                        if cat_slot_value_acc != "NAN" and cat_slot_value_acc != 1.0:
                            found_err = True
                        if noncat_slot_value_acc != "NAN" and noncat_slot_value_acc != 1.0:
                            found_err = True

                        if found_err:
                            logging.debug("-----------------------------------New Frame------------------------------")
                            logging.debug(
                                f'DIALOGUE ID : {dialog_id}, TURN ID: {turn_id}, SERVICE: {frame["service"]}, PREV_SERVICE: {frame_service_prev}'
                            )

                            logging.debug(f'SYS : {system_utterance}')
                            logging.debug(f'USER: {user_utterance}')

                            logging.debug("\n")
                            logging.debug(f"PRED CAT: {categorical_slots_dict}")
                            logging.debug(f"PRED NON-CAT: {non_categorical_slots_dict}")

                            logging.debug("\n")
                            logging.debug(f"STATE - LABEL: {sorted(true_state['slot_values'].items())}")
                            logging.debug(f"STATE - PRED : {sorted(slot_values.items())}")
                            logging.debug(f"STATE - PREV: {true_state_prev}")

                            logging.debug("\n")
                            logging.debug(f"SLOTS - LABEL: {true_slots}")
                            logging.debug(f"SYS SLOT AGG: {sys_slots_agg}")
                            logging.debug(f"SYS SLOT LAST: {sys_slots_last}")
                            logging.debug(f"SYS RETS: {sys_rets}")

                            logging.debug("\n")
                            logging.debug(f"CAT STATUS ACC: {cat_slot_status_acc}")

                            logging.debug(f"NONCAT STATUS ACC: {noncat_slot_status_acc}")

                            logging.debug(
                                f"CAT VALUES ACC: {cat_slot_value_acc} ,NONCAT VALUES ACC: {noncat_slot_value_acc}"
                            )

                            found_err = False
                            if cat_slot_status_acc != "NAN" and cat_slot_status_acc < 1.0:
                                logging.debug("CAT_STATUS_ERR")
                                found_err = True
                            if noncat_slot_status_acc != "NAN" and noncat_slot_status_acc < 1.0:
                                logging.debug("NONCAT_STATUS_ERR")
                                found_err = True
                            if (
                                noncat_slot_status_acc != "NAN"
                                and noncat_slot_status_acc < 1.0
                                and cat_slot_status_acc != "NAN"
                                and cat_slot_status_acc < 1.0
                            ):
                                logging.debug("BOTH_STATUS_ERR")
                                found_err = True

                            if cat_slot_value_acc != "NAN" and cat_slot_value_acc < 1.0:
                                logging.debug("CAT_VALUE_ERR")
                                found_err = True
                            if noncat_slot_value_acc != "NAN" and noncat_slot_value_acc < 1.0:
                                logging.debug("NONCAT_VALUE_ERR")
                                found_err = True
                            if (
                                noncat_slot_value_acc != "NAN"
                                and noncat_slot_value_acc != 1.0
                                and cat_slot_value_acc != "NAN"
                                and cat_slot_value_acc != 1.0
                            ):
                                logging.debug("BOTH_VALUE_ERR")
                                found_err = True
                            if not found_err:
                                logging.debug("CLEAN_FRAME")

                # Create a new dict to avoid overwriting the state in previous turns
                # because of use of same objects.
                state["slot_values"] = {s: [v] for s, v in slot_values.items()}
                frame["state"] = state
                frame_service_prev = frame["service"]

    return dialog


def get_predicted_dialog_baseline(dialog, all_predictions, schemas):
    """Update labels in a dialogue based on model predictions.
  Args:
    dialog: A json object containing dialogue whose labels are to be updated.
    all_predictions: A dict mapping prediction name to the predicted value. See
      SchemaGuidedDST class for the contents of this dict.
    schemas: A Schema object wrapping all the schemas for the dataset.
  Returns:
    A json object containing the dialogue with labels predicted by the model.
  """
    # Overwrite the labels in the turn with the predictions from the model. For
    # test set, these labels are missing from the data and hence they are added.
    dialog_id = dialog["dialogue_id"]
    # The slot values tracked for each service.
    all_slot_values = defaultdict(dict)
    for turn_idx, turn in enumerate(dialog["turns"]):
        if turn["speaker"] == "USER":
            user_utterance = turn["utterance"]
            system_utterance = dialog["turns"][turn_idx - 1]["utterance"] if turn_idx else ""
            turn_id = "{:02d}".format(turn_idx)
            for frame in turn["frames"]:
                predictions = all_predictions[(dialog_id, turn_id, frame["service"])]
                slot_values = all_slot_values[frame["service"]]
                service_schema = schemas.get_service_schema(frame["service"])
                # Remove the slot spans and state if present.
                frame.pop("slots", None)
                frame.pop("state", None)

                # The baseline model doesn't predict slot spans. Only state predictions
                # are added.
                state = {}

                # Add prediction for active intent. Offset is subtracted to account for
                # NONE intent.
                active_intent_id = predictions["intent_status"]
                state["active_intent"] = (
                    service_schema.get_intent_from_id(active_intent_id - 1) if active_intent_id else "NONE"
                )

                # Add prediction for requested slots.
                requested_slots = []
                for slot_idx, slot in enumerate(service_schema.slots):
                    if predictions["req_slot_status"][slot_idx] > REQ_SLOT_THRESHOLD:
                        requested_slots.append(slot)
                state["requested_slots"] = requested_slots

                # Add prediction for user goal (slot values).
                # Categorical slots.
                for slot_idx, slot in enumerate(service_schema.categorical_slots):
                    slot_status = predictions["cat_slot_status"][slot_idx]
                    if slot_status == STATUS_DONTCARE:
                        slot_values[slot] = STR_DONTCARE
                    elif slot_status == STATUS_ACTIVE:
                        value_idx = predictions["cat_slot_value"][slot_idx]
                        slot_values[slot] = service_schema.get_categorical_slot_values(slot)[value_idx]
                # Non-categorical slots.
                for slot_idx, slot in enumerate(service_schema.non_categorical_slots):
                    slot_status = predictions["noncat_slot_status"][slot_idx]
                    if slot_status == STATUS_DONTCARE:
                        slot_values[slot] = STR_DONTCARE
                    elif slot_status == STATUS_ACTIVE:
                        tok_start_idx = predictions["noncat_slot_start"][slot_idx]
                        tok_end_idx = predictions["noncat_slot_end"][slot_idx]
                        ch_start_idx = predictions["noncat_alignment_start"][tok_start_idx]
                        ch_end_idx = predictions["noncat_alignment_end"][tok_end_idx]
                        if ch_start_idx < 0 and ch_end_idx < 0:
                            # Add span from the system utterance.
                            slot_values[slot] = system_utterance[-ch_start_idx - 1 : -ch_end_idx]
                        elif ch_start_idx > 0 and ch_end_idx > 0:
                            # Add span from the user utterance.
                            slot_values[slot] = user_utterance[ch_start_idx - 1 : ch_end_idx]
                # Create a new dict to avoid overwriting the state in previous turns
                # because of use of same objects.
                state["slot_values"] = {s: [v] for s, v in slot_values.items()}
                frame["state"] = state
    return dialog


def write_predictions_to_file(
    predictions, input_json_files, output_dir, schemas, tracker_model, eval_debug, in_domain_services
):
    """Write the predicted dialogues as json files.

  Args:
    predictions: An iterator containing model predictions. This is the output of
      the predict method in the estimator.
    input_json_files: A list of json paths containing the dialogues to run
      inference on.
    schemas: Schemas to all services in the dst dataset (train, dev and test splits).
    output_dir: The directory where output json files will be created.
  """
    logging.info(f"Writing predictions to {output_dir} started.")

    # Index all predictions.
    all_predictions = {}
    for idx, prediction in enumerate(predictions):
        if not prediction["is_real_example"]:
            continue
        eval_dataset, dialog_id, turn_id, service_name = prediction['example_id'].split('-')
        all_predictions[(dialog_id, turn_id, service_name)] = prediction
    logging.info(f'Predictions for {idx} examples in {eval_dataset} dataset are getting processed.')

    # Read each input file and write its predictions.
    for input_file_path in input_json_files:
        with open(input_file_path) as f:
            dialogs = json.load(f)
            logging.debug(f'{input_file_path} file is loaded')
            pred_dialogs = []
            for d in dialogs:
                if tracker_model == 'baseline':
                    pred_dialog = get_predicted_dialog_baseline(d, all_predictions, schemas)
                elif tracker_model == 'nemotracker':
                    pred_dialog = get_predicted_dialog_nemotracker(
                        d, all_predictions, schemas, eval_debug, in_domain_services
                    )
                else:
                    raise ValueError(f"tracker_mode {tracker_model} is not defined.")
                pred_dialogs.append(pred_dialog)
            f.close()
        input_file_name = os.path.basename(input_file_path)
        output_file_path = os.path.join(output_dir, input_file_name)
        with open(output_file_path, "w") as f:
            json.dump(pred_dialogs, f, indent=2, separators=(",", ": "), sort_keys=True)
            f.close()



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2019 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os

import pytest
from ruamel.yaml import YAML

from nemo.collections.asr.models import ASRConvCTCModel


@pytest.mark.usefixtures("neural_factory")
class TestNeMoModels:
    @pytest.mark.unit
    def test_quartznet_creation(self):
        yaml = YAML(typ="safe")
        with open(
            os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../examples/asr/configs/jasper_an4.yaml"))
        ) as file:
            model_definition = yaml.load(file)
        model = ASRConvCTCModel(
            preprocessor_params=model_definition['AudioToMelSpectrogramPreprocessor'],
            encoder_params=model_definition['JasperEncoder'],
            decoder_params=model_definition['JasperDecoderForCTC'],
        )
        assert model.num_weights > 0
        assert len(model.modules) == 3

    @pytest.mark.unit
    def test_quartznet_nemo_file_export_and_import(self, tmpdir):
        yaml = YAML(typ="safe")
        with open(
            os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../examples/asr/configs/jasper_an4.yaml"))
        ) as file:
            model_definition = yaml.load(file)
        model = ASRConvCTCModel(
            preprocessor_params=model_definition['AudioToMelSpectrogramPreprocessor'],
            encoder_params=model_definition['JasperEncoder'],
            decoder_params=model_definition['JasperDecoderForCTC'],
        )
        nemo_file = str(tmpdir.mkdir("tmp_export_import").join("deleteme.nemo"))
        model.save_to(nemo_file)
        assert os.path.exists(nemo_file)
        new_qn = ASRConvCTCModel.from_pretrained(model_info=nemo_file)
        assert model.num_weights == new_qn.num_weights



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np
import torch

from nemo import logging
from nemo.collections.nlp.utils.callback_utils import tensor2list, tensor2numpy

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars, data_desc):

    if 'loss' not in global_vars:
        global_vars['loss'] = []
    if 'comp_res' not in global_vars:
        global_vars['comp_res'] = []
    if 'gating_labels' not in global_vars:
        global_vars['gating_labels'] = []
    if 'gating_preds' not in global_vars:
        global_vars['gating_preds'] = []

    point_outputs_max_list = []
    tgt_ids_list = []
    gate_outputs_max_list = []
    for tensor_name, values_list in tensors.items():
        if tensor_name.startswith('gating_labels'):
            for values in values_list:
                global_vars['gating_labels'].extend(tensor2list(values))
        elif tensor_name.startswith('point_outputs'):
            for values in values_list:
                p_max = torch.argmax(values, dim=-1)
                point_outputs_max_list.append(tensor2numpy(p_max))
        elif tensor_name.startswith('gate_outputs'):
            for values in values_list:
                g_max = torch.argmax(values, axis=-1)
                gate_outputs_max_list.append(tensor2numpy(g_max))
        elif tensor_name.startswith('tgt_ids'):
            for values in values_list:
                tgt_ids_list.append(tensor2numpy(values))

    comp_res_list = []
    for i in range(len(point_outputs_max_list)):
        mask_paddings = tgt_ids_list[i] == data_desc.vocab.pad_id
        comp_res = (point_outputs_max_list[i] == tgt_ids_list[i]) | mask_paddings
        comp_res = np.all(comp_res, axis=-1, keepdims=False)
        comp_res_list.extend(comp_res.tolist())

    gate_outputs_max = np.concatenate(gate_outputs_max_list, axis=0).tolist()

    global_vars['comp_res'].extend(comp_res_list)
    global_vars['gating_preds'].extend(gate_outputs_max)


def eval_epochs_done_callback(global_vars, data_desc):
    joint_acc, slot_acc = evaluate_metrics(
        global_vars['comp_res'],
        global_vars['gating_labels'],
        global_vars['gating_preds'],
        data_desc.gating_dict["ptr"],
    )

    gating_comp_flatten = (np.asarray(global_vars['gating_labels']) == np.asarray(global_vars['gating_preds'])).ravel()
    gating_acc = np.sum(gating_comp_flatten) / len(gating_comp_flatten)

    evaluation_metrics = {"Joint_Goal_Acc": joint_acc, "Slot_Acc": slot_acc, "Gate_Acc": gating_acc}
    logging.info(evaluation_metrics)

    return evaluation_metrics


def evaluate_metrics(comp_res, gating_labels, gating_preds, ptr_code):
    # TODO: Calculate precision, recall, and F1
    total_slots = 0
    correct_slots = 0
    total_turns = 0
    correct_turns = 0
    for result_idx, result in enumerate(comp_res):
        turn_wrong = False
        total_turns += 1
        for slot_idx, slot_eq in enumerate(result):
            total_slots += 1
            if gating_labels[result_idx][slot_idx] == ptr_code:
                if slot_eq:
                    correct_slots += 1
                else:
                    turn_wrong = True
            elif gating_labels[result_idx][slot_idx] == gating_preds[result_idx][slot_idx] or (
                slot_eq and gating_preds[result_idx][slot_idx] == ptr_code
            ):
                correct_slots += 1
            else:
                turn_wrong = True
        if not turn_wrong:
            correct_turns += 1

    slot_acc = correct_slots / float(total_slots) if total_slots != 0 else 0
    joint_acc = correct_turns / float(total_turns) if total_turns != 0 else 0
    return joint_acc, slot_acc



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# =============================================================================
# Copyright 2019 Salesforce Research.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# =============================================================================

import numpy as np
import torch
from torch.utils import data as pt_data

import nemo
from nemo.collections.nlp.data.datasets.multiwoz_dataset import MultiWOZDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core.neural_types import LabelsType, Length, NeuralType, TokenIndex
from nemo.utils.decorators import add_port_docs

__all__ = ['MultiWOZDataLayer']


class MultiWOZDataLayer(TextDataLayer):
    """
    Creates the data layer to use for State Tracking dataset MultiWOZ.

    Args:
        data_dir (str): path of the data folder
        domains (dict): dictionary of the domains to include
        all_domains (dict): dictionary of all the available domains
        vocab (Vocab): Vocabulary
        slots (list): list of the slots
        gating_dict (dict): dictionary of the gates
        num_samples (int): number of samples to include
        batch_size (int): batch size
        mode (str): mode of dataset, default='train'
        dataset_type (Dataset): Dataset Type
        shuffle (bool): enables shuffling, default=False
        num_workers (int): number of workers
        input_dropout (float): dropout applied to the input
        is_training (bool): specifies if it is for training
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        src_ids: ids of input sequences
        src_lens: lengths of input sequences
        tgt_ids: labels for the generator output
        tgt_lens: lengths of the generator targets
        gating_labels: labels for the gating head
        turn_domain: list of the domains

        """
        return {
            "src_ids": NeuralType(('B', 'T'), TokenIndex()),
            "src_lens": NeuralType(tuple('B'), Length()),
            "tgt_ids": NeuralType(('B', 'D', 'T'), LabelsType()),
            "tgt_lens": NeuralType(('B', 'D'), Length()),
            "gating_labels": NeuralType(('B', 'D'), LabelsType()),
            "turn_domain": NeuralType(),
        }

    def __init__(
        self,
        data_dir,
        domains,
        all_domains,
        vocab,
        slots,
        gating_dict,
        num_samples=-1,
        batch_size=16,
        mode='train',
        dataset_type=MultiWOZDataset,
        shuffle=False,
        num_workers=0,
        input_dropout=0,
        is_training=False,
    ):

        dataset_params = {
            'data_dir': data_dir,
            'domains': domains,
            'num_samples': num_samples,
            'mode': mode,
            'shuffle': shuffle,
            'all_domains': all_domains,
            'vocab': vocab,
            'slots': slots,
            'gating_dict': gating_dict,
        }
        super().__init__(dataset_type, dataset_params, batch_size=batch_size)

        if self._placement == nemo.core.DeviceType.AllGpu:
            sampler = pt_data.distributed.DistributedSampler(self._dataset)
        else:
            sampler = None

        self._dataloader = pt_data.DataLoader(
            dataset=self._dataset,
            batch_size=batch_size,
            shuffle=sampler is None,
            num_workers=num_workers,
            collate_fn=self._collate_fn,
            sampler=sampler,
        )
        self.pad_id = self._dataset.vocab.pad_id
        self.gating_dict = self._dataset.gating_dict
        self.input_dropout = input_dropout
        self.is_training = is_training
        self.vocab = self._dataset.vocab
        self.slots = self._dataset.slots

    def _collate_fn(self, data):
        """ data is a list of batch_size sample
        each sample is a dictionary of features
        """

        def pad_batch_context(sequences):
            '''
            merge from batch * sent_len to batch * max_len
            '''
            lengths = [len(seq) for seq in sequences]
            max_len = 1 if max(lengths) == 0 else max(lengths)
            for i, seq in enumerate(sequences):
                sequences[i] = seq + [1] * (max_len - len(seq))
            return torch.tensor(sequences), torch.tensor(lengths)

        def pad_batch_response(sequences, pad_id):
            '''
            merge from batch * nb_slot * slot_len to batch * nb_slot * max_slot_len
            '''
            lengths = []
            for bsz_seq in sequences:
                length = [len(v) for v in bsz_seq]
                lengths.append(length)
            max_len = max([max(l) for l in lengths])
            padded_seqs = []
            for bsz_seq in sequences:
                pad_seq = []
                for v in bsz_seq:
                    v = v + [pad_id] * (max_len - len(v))
                    pad_seq.append(v)
                padded_seqs.append(pad_seq)
            padded_seqs = torch.tensor(padded_seqs)
            lengths = torch.tensor(lengths)
            return padded_seqs, lengths

        data.sort(key=lambda x: len(x['context_ids']), reverse=True)
        item_info = {}
        for key in data[0]:
            item_info[key] = [item[key] for item in data]

        src_ids, src_lens = pad_batch_context(item_info['context_ids'])
        tgt_ids, tgt_lens = pad_batch_response(item_info['responses_ids'], self._dataset.vocab.pad_id)
        gating_label = torch.tensor(item_info['gating_label'])
        turn_domain = torch.tensor(item_info['turn_domain'])

        if self.input_dropout > 0 and self.is_training:
            bi_mask = np.random.binomial([np.ones(src_ids.size())], 1.0 - self.input_dropout)[0]
            rand_mask = torch.Tensor(bi_mask).long().to(src_ids.device)
            src_ids = src_ids * rand_mask

        return (
            src_ids.to(self._device),
            src_lens.to(self._device),
            tgt_ids.to(self._device),
            tgt_lens.to(self._device),
            gating_label.to(self._device),
            turn_domain.to(self._device),
        )

    @property
    def dataset(self):
        return None

    @property
    def data_iterator(self):
        return self._dataloader



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo import logging

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars):
    if "eval_start_logits" not in global_vars.keys():
        global_vars["eval_start_logits"] = []
    if "eval_end_logits" not in global_vars.keys():
        global_vars["eval_end_logits"] = []
    if "eval_unique_ids" not in global_vars.keys():
        global_vars["eval_unique_ids"] = []
    for kv, v in tensors.items():

        if 'logits' in kv:
            logits = []
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits.append(logit_tensor.detach().cpu().tolist())

            if kv.startswith('start_logits'):
                global_vars['eval_start_logits'].extend(logits)
            elif kv.startswith('end_logits'):
                global_vars['eval_end_logits'].extend(logits)

        if 'unique_ids' in kv:
            unique_ids = []
            for v_tensor in v:
                for id_tensor in v_tensor:
                    unique_ids.append(id_tensor.detach().cpu().tolist())
            global_vars['eval_unique_ids'].extend(unique_ids)


def eval_epochs_done_callback(
    global_vars,
    eval_data_layer,
    do_lower_case,
    n_best_size,
    max_answer_length,
    version_2_with_negative,
    null_score_diff_threshold,
):
    exact_match, f1, _, _ = eval_data_layer.dataset.evaluate(
        unique_ids=global_vars["eval_unique_ids"],
        start_logits=global_vars["eval_start_logits"],
        end_logits=global_vars["eval_end_logits"],
        n_best_size=n_best_size,
        max_answer_length=max_answer_length,
        version_2_with_negative=version_2_with_negative,
        null_score_diff_threshold=null_score_diff_threshold,
        do_lower_case=do_lower_case,
    )

    logging.info(f"Exact_match = {exact_match}, f1 = {f1}")

    global_vars["eval_unique_ids"] = []
    global_vars["eval_start_logits"] = []
    global_vars["eval_end_logits"] = []

    return dict({"exact_match": exact_match, "f1": f1})



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.nm.data_layers.bert_inference_datalayer import *
from nemo.collections.nlp.nm.data_layers.glue_benchmark_datalayer import *
from nemo.collections.nlp.nm.data_layers.joint_intent_slot_datalayer import *
from nemo.collections.nlp.nm.data_layers.lasertagger_datalayer import *
from nemo.collections.nlp.nm.data_layers.lm_bert_datalayer import *
from nemo.collections.nlp.nm.data_layers.lm_transformer_datalayer import *
from nemo.collections.nlp.nm.data_layers.machine_translation_datalayer import *
from nemo.collections.nlp.nm.data_layers.punctuation_capitalization_datalayer import *
from nemo.collections.nlp.nm.data_layers.qa_squad_datalayer import *
from nemo.collections.nlp.nm.data_layers.state_tracking_sgd_datalayer import *
from nemo.collections.nlp.nm.data_layers.state_tracking_trade_datalayer import *
from nemo.collections.nlp.nm.data_layers.text_classification_datalayer import *
from nemo.collections.nlp.nm.data_layers.text_datalayer import *
from nemo.collections.nlp.nm.data_layers.token_classification_datalayer import *



from itertools import chain
from pprint import pformat

import torch

from nemo.backends.pytorch.common.metrics import char_lm_metrics
from nemo.collections.asr.metrics import word_error_rate
from nemo.utils import logging

ENG_MWN = 5.3


def process_evaluation_batch(tensors, global_vars, labels, specials, tb_writer=None, write_attn=True):
    loss, log_probs = ([],) * 2
    transcripts, transcript_texts = ([],) * 2
    predictions, prediction_texts = ([],) * 2
    attention_weights = []
    for k, v in tensors.items():
        if 'loss' in k:
            loss = v
        elif 'log_probs' in k:
            log_probs = v
        elif ('transcripts' in k) or ('texts' in k):
            transcripts = v
            transcript_texts = __decode(v, labels, specials)
        elif 'predictions' in k:
            # predictions = v
            prediction_texts = __decode(v, labels, specials)
        elif 'attention_weights' in k:
            attention_weights = v

    global_vars.setdefault('loss', [])
    global_vars['loss'].extend(loss)
    bpc, ppl = char_lm_metrics(log_probs, transcripts, transcript_texts, specials['pad_id'])
    global_vars.setdefault('bpc', [])
    global_vars['bpc'].extend(bpc)
    global_vars.setdefault('ppl', [])
    global_vars['ppl'].extend(ppl)
    global_vars.setdefault('transcript_texts', [])
    global_vars['transcript_texts'].extend(transcript_texts)
    global_vars.setdefault('prediction_texts', [])
    global_vars['prediction_texts'].extend(prediction_texts)

    # TODO: Add step number?
    if tb_writer is not None and len(attention_weights) and write_attn:
        sample_len = len(prediction_texts[0][0])
        if sample_len > 0:
            attention_weights = attention_weights[0][0, :sample_len, :]
            tb_writer.add_image(
                'image/eval_attention_weights', attention_weights, dataformats='HW',
            )


def process_evaluation_epoch(
    global_vars, metrics=('loss', 'bpc', 'ppl'), calc_wer=False, mode='eval', tag='none',
):
    tag = '_'.join(tag.lower().strip().split())
    return_dict = {}
    for metric in metrics:
        value = torch.mean(torch.stack(global_vars[metric])).item()
        return_dict[f'metric/{mode}_{metric}_{tag}'] = value

    # TODO: Delete?
    bpc = return_dict[f'metric/{mode}_bpc_{tag}']
    return_dict[f'metric/{mode}_ppl_{tag}'] = 2 ** (bpc * ENG_MWN)

    if calc_wer:
        transcript_texts = list(chain(*global_vars['transcript_texts']))
        prediction_texts = list(chain(*global_vars['prediction_texts']))

        logging.info(f'Ten examples (transcripts and predictions)')
        logging.info(transcript_texts[:10])
        logging.info(prediction_texts[:10])

        wer = word_error_rate(hypotheses=prediction_texts, references=transcript_texts)
        return_dict[f'metric/{mode}_wer_{tag}'] = wer

    logging.info(pformat(return_dict))

    return return_dict


def __decode(tensors_list, labels, specials):
    labels_map = dict([(i, labels[i]) for i in range(len(labels)) if i not in set(specials.values())])
    results = []
    for tensor in tensors_list:
        tensor = tensor.long().cpu()
        hypotheses = []
        for i in range(tensor.shape[0]):
            hypothesis = ''.join([labels_map[c] for c in tensor[i].numpy().tolist() if c in labels_map])
            hypotheses.append(hypothesis)

        results.append(hypotheses)

    return results



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019-, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import shutil
import string
import tarfile
from abc import abstractmethod
from os import makedirs, path
from typing import Iterable

from nemo.core.neural_factory import DeploymentFormat, OperationMode
from nemo.core.neural_graph import NeuralGraph
from nemo.core.neural_modules import ModuleType, NeuralModule
from nemo.utils import logging

__all__ = ['NeMoModel']

NEMO_TMP_FOLDER = ".nemo_tmp"


class NeMoModel(NeuralModule):
    """Abstract class representing NeMoModel.
    A NeMoModel is a kind of neural module which contains other neural modules and logic inside.
    It typically represents a whole neural network and requires only connections with data layer and loss
    modules for training.

    The same NeMoModel could be used in training, evaluation or inference regimes. It should adjust itself
    accordingly.
    """

    def __call__(self, **kwargs):
        if self._operation_mode == OperationMode.training or self.operation_mode == OperationMode.both:
            return self.train_call(**kwargs)

        else:
            return self.eval_call(**kwargs)

    @abstractmethod
    def train_call(self, **kwargs):
        """Override this method with training logic"""
        pass

    @abstractmethod
    def eval_call(self, **kwargs):
        """Override this method with evaluation logic"""
        pass

    @classmethod
    @abstractmethod
    def from_pretrained(cls, model_info, local_rank: int = 0, referesh_cache: bool = False) -> NeuralModule:
        """
        Instantiates NeMoModel from pretrained checkpoint. Can do so from file on disk or from the NVIDIA NGC.
        Args:
            model_info: Either path to ".nemo" file or a valid NGC Model name
            local_rank: on which GPU to instantiate.
            referesh_cache: If set to True, then when fetching from clould, this will re-fetch the file
                from clould even if it is  already found in a cache locally.

        Returns:
            NeMoModel instance
        Raises:
            NotImplemened exception when there is no pre-trained models on the cloud
        """
        if isinstance(model_info, str) and model_info.endswith(".nemo"):
            nemo_file_folder, to_delete = cls.__unpack_nemo_file(path2file=model_info)
            configuration_file = path.join(nemo_file_folder, 'module.yaml')
            instance = cls.import_from_config(config_file=configuration_file)
            for module in instance.modules:
                module_checkpoint = path.join(nemo_file_folder, module.__class__.__name__ + ".pt")
                module.restore_from(path=module_checkpoint, local_rank=local_rank)
            shutil.rmtree(to_delete)
            return instance
        else:
            raise NotImplemented("Generic from_pretrained from cloud is not implemented")

    def save_to(self, output_file_name: str, output_folder: str = None, optimize_for_deployment: bool = False) -> str:
        """
        Saves NeMoModel to .nemo file. This file will contain:
            * weights of all NeuralModule instances inside the model
            * Yaml file with configuration
            * Yaml files with topologies and configuration of training and (if applicable) eval graphs
        Args:
            output_file_name: filename, something like nemomodel.nemo
            output_folder: folder where to save output_file_name. If None (default) current folder will be used.
            optimize_for_deployment: will optimize for deployment by trying to export modules to .onnx format and
                skipping training graph.

        Returns:
            None
        """

        def __make_nemo_file_from_folder(filename, source_dir):
            with tarfile.open(filename, "w:gz") as tar:
                tar.add(source_dir, arcname=path.basename(source_dir))

        if output_folder is None:
            output_folder = ""
        # create temporary folder first
        rnd_string = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(16))
        rnd_path = path.join(output_folder, f".{rnd_string}")
        tmp_folder = path.join(rnd_path, NEMO_TMP_FOLDER)
        # resulting_file is path to the resulting .nemo file
        if output_file_name.endswith(".nemo"):
            resulting_file = path.join(output_folder, output_file_name)
        else:
            resulting_file = path.join(output_folder, output_file_name + ".nemo")
        if not path.exists(tmp_folder):
            makedirs(tmp_folder)
        try:
            # create header file
            main_configuration_file_name = "module.yaml"
            train_graph_file_name = "train_graph.yaml"
            eval_graph_file_name = "eval_graph.yaml"

            header_file_path = path.join(tmp_folder, 'header.content')
            with open(header_file_path, 'w') as hf:
                hf.write(f"Main module configuration: {main_configuration_file_name}")
                hf.write(f"Train graph: {train_graph_file_name}")
                if self.eval_graph is not None:
                    hf.write(f"Eval graph: {train_graph_file_name}")

            # Every NeMo model is a NeuralModule, exporiting its hyperparamers to .yaml
            config_file_path = path.join(tmp_folder, main_configuration_file_name)
            self.export_to_config(config_file=config_file_path)

            if self.train_graph is not None and not optimize_for_deployment:
                config_file_path_train_graph = path.join(tmp_folder, train_graph_file_name)
                self.train_graph.export_to_config(config_file_path_train_graph)

            if self.eval_graph is not None:
                config_file_path_eval_graph = path.join(tmp_folder, eval_graph_file_name)
                self.eval_graph.export_to_config(config_file_path_eval_graph)

            for module in self.modules:
                module_name = module.__class__.__name__
                if optimize_for_deployment:
                    try:
                        module_checkpoint = module_name + ".onnx"
                        module._factory.deployment_export(
                            module=module,
                            output=path.join(tmp_folder, module_checkpoint),
                            d_format=DeploymentFormat.TRTONNX,
                        )
                    except Exception as ex:
                        logging.warning(ex)
                        logging.warning(f"Did not convert {module_name} to .onnx")
                        module_checkpoint = module_name + ".pt"
                        module.save_to(path.join(tmp_folder, module_checkpoint))
                else:
                    module_checkpoint = module_name + ".pt"
                    module.save_to(path.join(tmp_folder, module_checkpoint))

            __make_nemo_file_from_folder(resulting_file, tmp_folder)
            logging.info(f"Exported model {self} to {resulting_file}")
        except Exception as ex:
            logging.error(ex)
            logging.error("Could not perform NeMoModel export")
        finally:
            shutil.rmtree(rnd_path)

    @property
    @abstractmethod
    def modules(self) -> Iterable[NeuralModule]:
        pass

    @property
    def train_graph(self) -> NeuralGraph:
        return None

    @property
    def eval_graph(self) -> NeuralGraph:
        return None

    def train(self):
        """
        Sets model to the training mode

        Returns:
            None
        """
        self._operation_mode = OperationMode.training
        for module in self.modules:
            module.operation_mode = OperationMode.training
            if module.type == ModuleType.trainable and hasattr(module, 'train'):
                module.train()

    def eval(self):
        """
        Sets model to the evaluation mode

        Returns:
            None
        """
        self._operation_mode = OperationMode.evaluation
        for module in self.modules:
            module.operation_mode = OperationMode.evaluation
            if module.type == ModuleType.trainable and hasattr(module, 'eval'):
                module.eval()

    @staticmethod
    def __unpack_nemo_file(path2file: str, out_folder: str = None) -> str:
        if not path.exists(path2file):
            raise FileNotFoundError(f"{path2file} does not exist")
        if out_folder is None:
            out_folder = ''.join(
                random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(16)
            )

        tar = tarfile.open(path2file, "r:gz")
        tar.extractall(path=out_folder)
        tar.close()
        return path.join(out_folder, NEMO_TMP_FOLDER), out_folder

    def get_weights(self):
        raise NotImplementedError()

    def set_weights(
        self, name2weight, name2name_and_transform,
    ):
        raise NotImplementedError()

    def tie_weights_with(
        self, module, weight_names, name2name_and_transform,
    ):
        raise NotImplementedError()

    def restore_from(self, path: str):
        raise NotImplementedError("Please use from_pretrained method for NeMoModels")

    def freeze(self, weights):
        raise NotImplementedError()

    def unfreeze(self, weights):
        raise NotImplementedError()



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import numpy as np

from nemo import logging

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars):

    for kv, v in tensors.items():

        if 'smoothedcrossentropyloss' in kv:
            if "dev_mlm_loss" not in global_vars.keys():
                global_vars["dev_mlm_loss"] = []
            for dev_mlm_loss in v:
                global_vars["dev_mlm_loss"].append(dev_mlm_loss.item())
        if 'crossentropylossnm' in kv:
            if "dev_nsp_loss" not in global_vars.keys():
                global_vars["dev_nsp_loss"] = []
            for dev_nsp_loss in v:
                global_vars["dev_nsp_loss"].append(dev_nsp_loss.item())
        if 'lossaggregator' in kv:
            if "dev_loss" not in global_vars.keys():
                global_vars["dev_loss"] = []
            for dev_loss in v:
                global_vars["dev_loss"].append(dev_loss.item())


def eval_epochs_done_callback(global_vars):
    res = {}
    if 'dev_mlm_loss' in global_vars:
        mlm_loss = np.mean(global_vars["dev_mlm_loss"])
        logging.info("Dev MLM perplexity: {0}".format(np.round(np.exp(mlm_loss), 3)))
        global_vars["dev_mlm_loss"] = []
        res["Dev MLM loss"] = mlm_loss
    else:
        mlm_loss = -123.0

    if 'dev_nsp_loss' in global_vars:
        nsp_loss = np.mean(global_vars["dev_nsp_loss"])
        logging.info("Dev NSP perplexity: {0}".format(np.round(np.exp(nsp_loss), 3)))
        global_vars["dev_nsp_loss"] = []
        res["Dev NSP loss"] = nsp_loss
    else:
        nsp_loss = -123.0

    if 'dev_loss' in global_vars:
        total_loss = np.mean(global_vars["dev_loss"])
        logging.info("Dev perplexity: {0}".format(np.round(np.exp(total_loss), 3)))
        global_vars["dev_loss"] = []
        res["Dev loss"] = total_loss
    else:
        nsp_loss = -123.0

    return res



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
# Copyright (C) tkornuta, IBM Corporation 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__author__ = "Tomasz Kornuta"

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/IBM/pytorchpipe/blob/develop/ptp/components/transforms/reshape_tensor.py
"""

from typing import List, Optional

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.core.neural_types import AxisKind, AxisType, NeuralType, VoidType
from nemo.utils import logging
from nemo.utils.configuration_error import ConfigurationError
from nemo.utils.decorators import add_port_docs

__all__ = ['ReshapeTensor']


class ReshapeTensor(NonTrainableNM):
    """
    Class responsible for reshaping the input tensor.

    Reshapes tensor from e.g. [64, 16, 2, 2] to [64, 64].

    For more details please refer to: https://pytorch.org/docs/master/generated/torch.reshape.html
    """

    def __init__(self, input_sizes: List[int], output_sizes: List[int], name: Optional[str] = None):
        """
        Initializes the object.

        Args:
            input_sizes: Sizes of dimensions of the input tensor.
            output_sizes: Sizes of dimensions of the output.
            name: Name of the module (DEFAULT: None)
        """
        # Call constructor of parent classes.
        NonTrainableNM.__init__(self, name=name)

        # Validate params.
        if type(input_sizes) != list or len(input_sizes) < 2:
            raise ConfigurationError(
                "'input_sizes' must be at least a list with two values (received {})".format(self.input_sizes)
            )
        if type(output_sizes) != list or len(output_sizes) < 2:
            raise ConfigurationError(
                "'output_sizes' must be at least a list with two values (received {})".format(self.output_sizes)
            )

        # Get input and output shapes from configuration.
        self._input_sizes = input_sizes
        self._output_sizes = output_sizes

    @property
    @add_port_docs()
    def input_ports(self):
        """
        Returns definitions of module input ports.
        Batch of inputs, each represented as index [BATCH_SIZE x ... x INPUT_SIZE]
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        for size in self._input_sizes[1:]:
            axes.append(AxisType(kind=AxisKind.Any, size=size))
        # Return neural type.
        return {"inputs": NeuralType(axes, VoidType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """
        Returns definitions of module output ports.
        """
        # Prepare list of axes.
        axes = [AxisType(kind=AxisKind.Batch)]
        for size in self._output_sizes[1:]:
            axes.append(AxisType(kind=AxisKind.Any, size=size))
        # Return neural type.
        return {"outputs": NeuralType(axes, VoidType())}

    def forward(self, inputs):
        """
        Encodes "inputs" in the format of a single tensor.
        Stores reshaped tensor in "outputs" field of in data_streams.

        Args:
            inputs: a tensor [BATCH_SIZE x ...]

        Returns:
            Outputs a tensor [BATCH_SIZE x ...] 
        """
        # print("{}: input shape: {}, device: {}\n".format(self.name, inputs.shape, inputs.device))

        # Reshape.
        return inputs.view(self._output_sizes)
